"""
This is step 4/4 in predict.py

SeeAlso:

    predict.py

    prepare_kwcoco.py

    tile_processing_kwcoco.py

    export_cold_result_kwcoco.py

    assemble_cold_result_kwcoco.py *
"""
import os
import numpy as np
import pandas as pd
from osgeo import gdal
import pytz
import datetime as datetime_mod
from datetime import datetime as datetime_cls
import json
import kwcoco
import kwimage
import scriptconfig as scfg
import ubelt as ub
import logging
import gc
import shutil
try:
    from line_profiler import profile
except ImportError:
    from ubelt import identity as profile
from kwutil import util_time

logger = logging.getLogger(__name__)
# FIXME: note for Jon to be more general than assuming US/East locality in the future
tz = pytz.timezone('US/Eastern')


class AssembleColdKwcocoConfig(scfg.DataConfig):
    """
    TODO: write docs
    """
    stack_path = scfg.Value(None, help='folder directory of stacked data')
    reccg_path = scfg.Value(None, help='folder directory of cold processing result')
    coco_fpath = scfg.Value(None, help='file path of coco json')
    combined_coco_fpath = scfg.Value(None, help=ub.paragraph(
        '''
        a path to a file to combined kwcoco file
        '''))
    mod_coco_fpath = scfg.Value(None, help='file path for modified coco json')
    write_kwcoco = scfg.Value(True, help='writing kwcoco file based on COLD feature, Default is False')
    year_lowbound = scfg.Value(None, help='min year for saving geotiff, e.g., 2017')
    year_highbound = scfg.Value(None, help='max year for saving geotiff, e.g., 2022')
    coefs = scfg.Value(None, type=str, help="list of COLD coefficients for saving geotiff, e.g., a0,c1,a1,b1,a2,b2,a3,b3,cv,rmse")
    coefs_bands = scfg.Value(None, type=str, help='indicate the bands for output coefs_bands, e.g., 0,1,2,3,4,5')
    timestamp = scfg.Value(False, help='True: exporting cold result by timestamp, False: exporting cold result by year, Default is False')
    combine = scfg.Value(True, help='for temporal combined mode, Default is True')
    exclude_first = scfg.Value(True, help='exclude first date of image from each sensor, Default is True')
    resolution = scfg.Value('30GSD', help=ub.paragraph(
        '''
        the resolution used when preparing the kwcoco data. Note: results will
        be wrong if this does not agree with what was used in
        PrepareKwcocoConfig
        '''))
    sensors = scfg.Value('L8', type=str, help='sensor type, default is "L8"')
    cold_time_span = scfg.Value('1year', type=str, help='Temporal period for extracting cold features, default is "1year", another option is "6months"')


@profile
def assemble_main(cmdline=1, **kwargs):
    """_summary_

    Args:
        cmdline (int, optional): _description_. Defaults to 1.

    Ignore:
        python -m geowatch.tasks.cold.assemble_cold_result_kwcoco --help
        TEST_COLD=1 xdoctest -m geowatch.tasks.cold.assemble_cold_result_kwcoco assemble_main

    Example:
    >>> # xdoctest: +REQUIRES(env:TEST_COLD)
    >>> from geowatch.tasks.cold.assemble_cold_result_kwcoco import assemble_main
    >>> from geowatch.tasks.cold.assemble_cold_result_kwcoco import *
    >>> kwargs= dict(
    >>>    stack_path = "/gpfs/scratchfs1/zhz18039/jws18003/new-repos/smart_data_dvc2/Drop6-MeanYear10GSD-V2/_pycold/stacked/KR_R001/",
    >>>    reccg_path = "/gpfs/scratchfs1/zhz18039/jws18003/new-repos/smart_data_dvc2/Drop6-MeanYear10GSD-V2/_pycold/reccg/KR_R001/",
    >>>    coco_fpath = ub.Path('/gpfs/scratchfs1/zhz18039/jws18003/new-repos/smart_data_dvc2/Drop6/imgonly-KR_R001.kwcoco.json'),
    >>>    mod_coco_fpath = ub.Path('/gpfs/scratchfs1/zhz18039/jws18003/new-repos/smart_data_dvc2/Drop6/imgonly_KR_R001_cold.kwcoco.zip'),
    >>>    combined_coco_fpath = ub.Path('/gpfs/scratchfs1/zhz18039/jws18003/new-repos/smart_data_dvc2/Drop6-MeanYear10GSD-V2/imgonly-KR_R001.kwcoco.zip'),
    >>>    coefs = 'cv,rmse,a0,a1,b1,c1',
    >>>    year_lowbound = None,
    >>>    year_highbound = None,
    >>>    coefs_bands = '0,1,2,3,4,5',
    >>>    timestamp = False,
    >>>    combine = True,
    >>>    sensors = 'L8',
    >>>    resolution = '10GSD',
    >>>    )
    >>> cmdline=0
    >>> assemble_main(cmdline, **kwargs)
    """
    # a hacky way to pass the process context and progress manager from the
    # caller when this is called as a subroutine
    pman = kwargs.pop('pman', None)
    proc_context = kwargs.pop('proc_context', None)

    config_in = AssembleColdKwcocoConfig.cli(cmdline=cmdline, data=kwargs)
    stack_path = ub.Path(config_in['stack_path'])
    reccg_path = ub.Path(config_in['reccg_path'])
    coco_fpath = ub.Path(config_in['coco_fpath'])
    mod_coco_fpath = config_in['mod_coco_fpath']
    if mod_coco_fpath is not None:
        mod_coco_fpath = ub.Path(config_in['mod_coco_fpath'])
    write_kwcoco = config_in['write_kwcoco']
    if mod_coco_fpath is None and write_kwcoco is True:
        raise ValueError('Must specify mod_coco_fpath if write_kwcoco is True')
    out_path = reccg_path / 'cold_feature'
    tmp_path = out_path / 'tmp'
    year_lowbound = config_in['year_lowbound']
    year_highbound = config_in['year_highbound']
    coefs = config_in['coefs']
    coefs_bands = config_in['coefs_bands']
    timestamp = config_in['timestamp']
    combine = config_in['combine']
    exclude_first = config_in['exclude_first']
    resolution = config_in['resolution']
    sensors = config_in['sensors']
    cold_time_span = config_in['cold_time_span']
    cold_time_span = util_time.timedelta.coerce(cold_time_span)

    if config_in['combined_coco_fpath'] is not None:
        combined_coco_fpath = ub.Path(config_in['combined_coco_fpath'])
    else:
        if combine:
            raise ValueError('Must specify combined_coco_fpath if combine is True')
        combined_coco_fpath = None

    # define variables
    # config = read_json_metadata(stack_path)
    log_fpath = reccg_path / 'log.json'
    with open(log_fpath, "r") as f:
        config = json.load(f)
    vid_w = config['video_w']
    vid_h = config['video_h']
    n_block_x = config['n_block_x']
    n_block_y = config['n_block_y']
    n_blocks = n_block_x * n_block_y  # total number of blocks

    cold_param = json.loads((reccg_path / 'log.json').read_text())
    method = cold_param['algorithm']

    coef_names = ['cv', 'rmse', 'a0', 'a1', 'b1', 'a2', 'b2', 'a3', 'b3', 'c1']
    band_names = [0, 1, 2, 3, 4, 5]

    BAND_INFO = {0: 'blue',
                 1: 'green',
                 2: 'red',
                 3: 'nir',
                 4: 'swir16',
                 5: 'swir22'}

    if coefs is not None:
        try:
            coefs = list(coefs.split(","))
        except Exception:
            print("Illegal coefs inputs: example, --coefs='a0, c1, a1, b1, a2, b2, a3, b3, cv, rmse'")

        try:
            coefs_bands = list(coefs_bands.split(","))
            coefs_bands = [int(coefs_band) for coefs_band in coefs_bands]
        except Exception:
            print("Illegal coefs_bands inputs: example, --coefs_bands='0, 1, 2, 3, 4, 5, 6'")

    # Get original transform from projection to image space
    coco_dset = kwcoco.CocoDataset(coco_fpath)
    L8_new_gdal_transform, L8_proj = get_gdal_transform(coco_dset, 'L8', resolution=resolution)
    S2_new_gdal_transform, S2_proj = get_gdal_transform(coco_dset, 'S2', resolution=resolution)

    available_transforms = [L8_new_gdal_transform, S2_new_gdal_transform]
    if all(t is None for t in available_transforms):
        raise RuntimeError('There are no images of known sensors')

    # Define sensor-specific information
    # TODO: planetscope
    sensor_info = {
        'L8': {
            'outdriver': gdal.GetDriverByName('GTiff'),
            'new_gdal_transform': L8_new_gdal_transform,
            'proj': L8_proj
        },
        'S2': {
            'outdriver': gdal.GetDriverByName('GTiff'),
            'new_gdal_transform': S2_new_gdal_transform,
            'proj': S2_proj
        }
    }

    # Get ordinal day list
    block_folder = stack_path / 'block_x1_y1'

    # if timestamp:
    meta_files = [m for m in os.listdir(block_folder) if m.endswith('.json')]

    # Create dictionaries to store ordinal dates and image names for each sensor
    ordinal_dates = {}
    img_names = {}
    sensors = list(sensors.split(","))

    # Initialize dictionaries for each sensor
    for sensor in sensors:
        ordinal_dates[sensor] = []
        img_names[sensor] = []

    # Read metadata and populate dictionaries
    for meta in meta_files:
        meta_config = json.loads((block_folder / meta).read_text())
        ordinal_date = meta_config['ordinal_date']
        img_name = meta_config['image_name'] + '.npy'
        for sensor in sensors:
            if f'_{sensor}_' in meta_config['image_name']:
                ordinal_dates[sensor].append(ordinal_date)
                img_names[sensor].append(img_name)
                break

    if year_lowbound is None:
        year_low_ordinal = min(min(ordinal_dates[sensor]) for sensor in sensors)
        year_lowbound = pd.Timestamp.fromordinal(year_low_ordinal).year
    else:
        year_low_ordinal = pd.Timestamp.toordinal(
            datetime_mod.datetime(int(year_lowbound), 1, 1))

    if year_highbound is None:
        year_high_ordinal = max(max(ordinal_dates[sensor]) for sensor in sensors)
        year_highbound = pd.Timestamp.fromordinal(year_high_ordinal).year
    else:
        year_high_ordinal = pd.Timestamp.toordinal(
            datetime_mod.datetime(int(year_highbound + 1), 1, 1))

    # Filter and sort img_dates and img_names based on the year bounds
    filtered_img_dates = {}
    filtered_img_names = {}
    for sensor in sensors:
        filtered_img_dates[sensor] = []
        filtered_img_names[sensor] = []

    for sensor in sensors:
        for date, name in zip(ordinal_dates[sensor], img_names[sensor]):
            if year_low_ordinal <= date < year_high_ordinal:
                filtered_img_dates[sensor].append(date)
                filtered_img_names[sensor].append(name)
        # Sort filtered img_dates
        filtered_img_dates[sensor] = sorted(filtered_img_dates[sensor])
        filtered_img_names[sensor] = sorted(filtered_img_names[sensor])

    if timestamp:
        img_dates = [date for sensor_dates in filtered_img_dates.values() for date in sensor_dates]
        ordinal_day_list = img_dates
    else:
        for sensor in sensors:
            year_group = {}
            img_name_group = {}
            ordinal_dates[sensor] = []
            img_names[sensor] = []
            for ordinal_day, img_name in zip(filtered_img_dates[sensor], filtered_img_names[sensor]):
                year = pd.Timestamp.fromordinal(ordinal_day).year
                if year not in year_group:
                    year_group[year] = []
                    img_name_group[year] = []
                year_group[year].append(ordinal_day)
                img_name_group[year].append(img_name)
            for year in sorted(year_group.keys()):
                year_group_by_year = year_group[year]
                # Determine the number of subdivisions
                num_subdivisions = int(365 / cold_time_span.days)
                # Select the first index from each subdivision
                for i in range(num_subdivisions):
                    # Calculate the start and end indices for the subdivision
                    start_idx = i * int(len(year_group_by_year) / num_subdivisions)
                    if start_idx < len(year_group_by_year):
                        ordinal_dates[sensor].append(year_group_by_year[start_idx])
                        img_names[sensor].append(img_name_group[year][start_idx])
        if exclude_first:
            ordinal_day_list = [date for _, dates in ordinal_dates.items() for date in dates[1:]]
            img_names_list = [name for _, names in img_names.items() for name in names[1:]]
        else:
            ordinal_day_list = [date for _, dates in ordinal_dates.items() for date in dates]
            img_names_list = [name for _, names in img_names.items() for name in names]
    if combine:
        combined_coco_dset = kwcoco.CocoDataset(combined_coco_fpath)

        # filter by sensors
        all_images = combined_coco_dset.images(list(ub.flatten(combined_coco_dset.videos().images)))
        flags = [s in sensors for s in all_images.lookup('sensor_coarse')]
        all_images = all_images.compress(flags)
        image_id_iter = iter(all_images)

        # Get ordinal date of combined coco image
        ordinal_dates = []
        img_names_list = []
        for image_id in image_id_iter:
            combined_coco_image: kwcoco.CocoImage = combined_coco_dset.coco_image(image_id)
            coco_image: kwcoco.CocoImage = coco_dset.coco_image(image_id)
            ts = combined_coco_image.img['timestamp']
            coco_img_name = coco_image.img['name']
            timestamp_local = datetime_cls.fromtimestamp(ts, tz=tz)
            timestamp_utc = timestamp_local.astimezone(pytz.utc)
            ordinal = timestamp_utc.toordinal()
            ordinal_dates.append(ordinal)
            img_names_list.append(coco_img_name)
        ordinal_day_list = ordinal_dates

    # assemble
    logger.info('Generating COLD output geotiff')
    if coefs is not None:
        day_iter = range(len(ordinal_day_list))
        if pman is not None:
            day_iter = pman.progiter(day_iter, total=len(ordinal_day_list))
        for day in day_iter:
            tmp_map_blocks = [np.load(
                tmp_path / f'tmp_coefmap_block{x + 1}_{ordinal_day_list[day]}.npy')
                for x in range(n_blocks)]

            results = np.hstack(tmp_map_blocks)
            results = np.vstack(np.hsplit(results, n_block_x))
            ninput = 0
            for band_idx, band_name in enumerate(coefs_bands):
                for coef_index, coef in enumerate(coefs):
                    kwcoco_img_name = os.path.splitext(img_names_list[day])[0]
                    band = BAND_INFO[band_name]
                    name_parts = list(map(str, (kwcoco_img_name, band, method, coef)))
                    outname = '_'.join(name_parts) + '.tif'
                    outfile = out_path / outname
                    for sensor, info in sensor_info.items():
                        if f'_{sensor}_' in outname:
                            outdriver = info['outdriver']
                            outdata = outdriver.Create(os.fspath(outfile), vid_w, vid_h, 1, gdal.GDT_Float32)
                            outdata.GetRasterBand(1).WriteArray(results[:vid_h, :vid_w, ninput])
                            outdata.GetRasterBand(1).SetNoDataValue(-9999)
                            outdata.FlushCache()
                            outdata.SetGeoTransform(info['new_gdal_transform'])
                            outdata.FlushCache()
                            outdata.SetProjection(info['proj'])
                            outdata.FlushCache()
                            ninput = ninput + 1

            # for x in range(n_blocks):
                # TODO: would be nice to have a structure that controls these
                # name formats so we can use padded inter suffixes for nicer
                # sorting, or nest files to keep folder sizes small

    # Remove tmp files
    shutil.rmtree(tmp_path)

    # logger.info('Starting adding new asset to kwcoco json')
    if write_kwcoco:
        if combine:
            output_coco_dset = combined_coco_dset
            for image_id in output_coco_dset.images():
                combined_coco_image: kwcoco.CocoImage = output_coco_dset.coco_image(image_id)
                coco_image: kwcoco.CocoImage = coco_dset.coco_image(image_id)
                image_name = coco_image.img['name']

                asset_w = vid_w
                asset_h = vid_h

                for band_name in band_names:
                    for coef in coef_names:
                        band = BAND_INFO[band_name]
                        new_fpath = out_path / f'{image_name}_{band}_{method}_{coef}.tif'
                        if new_fpath.exists():
                            channels = kwcoco.ChannelSpec.coerce(f'{band}_{method}_{coef}')
                            # COLD output was wrote based on transform information of
                            # coco_dset, so it aligned to a scaled video space.
                            warp_img_from_vid = combined_coco_image.warp_img_from_vid

                            if resolution is None:
                                scale_asset_from_vid = (1., 1.)
                            else:
                                scale_asset_from_vid = combined_coco_image._scalefactor_for_resolution(
                                    space='video', resolution=resolution)
                            warp_asset_from_vid = kwimage.Affine.scale(scale_asset_from_vid)
                            warp_vid_from_asset = warp_asset_from_vid.inv()
                            warp_img_from_asset = warp_img_from_vid @ warp_vid_from_asset

                            # Use the CocoImage helper which will augment the coco dictionary with
                            # your information.
                            combined_coco_image.add_asset(
                                file_name=new_fpath,
                                channels=channels,
                                width=asset_w,
                                height=asset_h,
                                warp_aux_to_img=warp_img_from_asset)
                            print(f'Added to the asset {new_fpath}')
                            logger.info(f'Added to the asset {new_fpath}')
        else:
            output_coco_dset = coco_dset
            for image_id in output_coco_dset.images():
                # Create a CocoImage object for each image.
                coco_image: kwcoco.CocoImage = output_coco_dset.coco_image(image_id)
                image_name = coco_image.img['name']

                asset_w = vid_w
                asset_h = vid_h

                for band_name in band_names:
                    for coef in coef_names:
                        band = BAND_INFO[band_name]
                        new_fpath = out_path / f'{image_name}_{band}_{method}_{coef}.tif'
                        if new_fpath.exists():
                            channels = kwcoco.ChannelSpec.coerce(f'{band}_{method}_{coef}')

                            # COLD output was wrote based on transform information of
                            # coco_dset, so it aligned to a scaled video space.
                            warp_img_from_vid = coco_image.warp_img_from_vid

                            if resolution is None:
                                scale_asset_from_vid = (1., 1.)
                            else:
                                scale_asset_from_vid = coco_image._scalefactor_for_resolution(
                                    space='video', resolution=resolution)
                            warp_asset_from_vid = kwimage.Affine.scale(scale_asset_from_vid)
                            warp_vid_from_asset = warp_asset_from_vid.inv()
                            warp_img_from_asset = warp_img_from_vid @ warp_vid_from_asset

                            # Use the CocoImage helper which will augment the coco dictionary with
                            # your information.
                            coco_image.add_asset(os.fspath(new_fpath),
                                                    channels=channels, width=asset_w,
                                                    height=asset_h,
                                                    warp_aux_to_img=warp_img_from_asset)

                            logger.info(f'Added to the asset {new_fpath}')

        if proc_context is not None:
            context_info = proc_context.stop()
            output_coco_dset.dataset['info'].append(context_info)

        # Write a modified kwcoco.json file
        logger.info(f'Writing kwcoco file to: {mod_coco_fpath}')
        if combine:
            combined_coco_dset.fpath = mod_coco_fpath
            combined_coco_dset._ensure_json_serializable()
            combined_coco_dset.dump()
        else:
            coco_dset.fpath = mod_coco_fpath
            coco_dset._ensure_json_serializable()
            coco_dset.dump()
        logger.info(f'Finished writing kwcoco file to: {mod_coco_fpath}')

    else:
        if proc_context is not None:
            context_info = proc_context.stop()
            coco_dset.dataset['info'].append(context_info)
    gc.collect()


@profile
def get_gdal_transform(coco_dset, sensor_name, resolution=None):
    video_ids = list(coco_dset.videos())
    if len(video_ids) != 1:
        raise AssertionError('currently expecting one video per coco file; todo: be robust to this')
    video_id = video_ids[0]

    # Get all the images in the video.
    images = coco_dset.images(video_id=video_id)
    sensors = images.lookup('sensor_coarse', None)
    is_target_sensor = [s == sensor_name for s in sensors]

    # Filter to only the images from target sensor
    target_images = images.compress(is_target_sensor)
    if len(target_images) == 0:
        return None, None
        raise RuntimeError(f'Video {video_id} in {coco_dset} contains no {sensor_name} images')

    # Take the first target image
    target_coco_img = target_images.coco_images[0]

    # Get the transform for the original asset
    target_primary_asset = target_coco_img.primary_asset()
    target_primary_fpath = os.path.join(ub.Path(target_coco_img.bundle_dpath), target_primary_asset['file_name'])
    try:
        ref_image = gdal.Open(target_primary_fpath, gdal.GA_ReadOnly)
        proj = ref_image.GetProjection()     # This transforms from world space to CRS84
        trans = ref_image.GetGeoTransform()  # This transforms from the underlying asset to world space.
    except AttributeError:
        return None, None
    warp_wld_from_primary = kwimage.Affine.from_gdal(trans)
    warp_img_from_primary = kwimage.Affine.coerce(target_primary_asset['warp_aux_to_img'])

    # If we requested a specific processing resolution, our *new* asset on disk
    # will differ by a scale factor.
    if resolution is None:
        scale_asset_from_vid = (1.0, 1.0)
    else:
        scale_asset_from_vid = target_coco_img._scalefactor_for_resolution(
            space='video', resolution=resolution)
    warp_asset_from_vid = kwimage.Affine.scale(scale_asset_from_vid)
    warp_img_from_vid = target_coco_img.warp_img_from_vid

    warp_vid_from_asset = warp_asset_from_vid.inv()
    warp_primary_from_img = warp_img_from_primary.inv()

    #
    # Calculate the new GDAL transform mapping our new asset on disk to the
    # world space
    warp_wld_from_asset = (
        warp_wld_from_primary @ warp_primary_from_img @
        warp_img_from_vid @ warp_vid_from_asset
    )
    new_geotrans = tuple(warp_wld_from_asset.to_gdal())
    return new_geotrans, proj


@profile
def read_json_metadata(stacked_path):
    for root, dirs, files in os.walk(stacked_path):
        for file in files:
            if file.endswith(".json"):
                json_path = os.path.join(root, file)

                with open(json_path, "r") as f:
                    metadata = json.load(f)
                    return metadata


if __name__ == '__main__':
    assemble_main()



# coding=utf-8
# Copyright 2022 The Deeplab2 Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This file contains code to build MaX-DeepLab output heads.

Reference:
  MaX-DeepLab: "End-to-End Panoptic Segmentation with Mask Transformers",
    CVPR 2021. https://arxiv.org/abs/2012.00759
      Huiyu Wang, Yukun Zhu, Hartwig Adam, Alan Yuille, Liang-Chieh Chen.
"""
import math

import tensorflow as tf
from deeplab2 import common
from deeplab2.model.decoder import panoptic_deeplab
from deeplab2.model.layers import convolutions

_PIXEL_SPACE_FEATURE_KEY = 'pixel_space_feature'


def _get_transformer_class_head_num_classes(
        auxiliary_semantic_head_output_channels,
        ignore_label):
    """Computes the num of classes for the transformer class head.

    The transformer class head predicts non-void classes (i.e., thing classes and
    stuff classes) and a void (i.e., ∅, no object) class. If the auxiliary
    semantic head output channel includes the void class, e.g., on COCO, we
    directly use the semantic output channel. Otherwise, e.g., on Cityscapes, we
    add 1 (the void class) to the transformer class head.

    Args:
      auxiliary_semantic_head_output_channels: An integer, the number of output
        channels of the auxiliary semantic head (it should be the same as the
        num_classes field of the dataset information).
      ignore_label: An integer specifying the ignore label. Default to 255.

    Returns:
      num_classes: An integer, the num of classes for the transformer class head.
    """
    if ignore_label >= auxiliary_semantic_head_output_channels:
        return auxiliary_semantic_head_output_channels + 1
    else:
        return auxiliary_semantic_head_output_channels


def add_bias_towards_void(transformer_class_logits, void_prior_prob=0.9):
    """Adds init bias towards the void (no object) class to the class logits.

    We initialize the void class with a large probability, similar to Section 3.3
    of the Focal Loss paper.

    Reference:
      Focal Loss for Dense Object Detection, ICCV 2017.
        https://arxiv.org/abs/1708.02002
          Tsung-Yi Lin, Priya Goyal, Ross Girshick, Kaiming He, Piotr Dollár.

    Args:
      transformer_class_logits: A [batch, num_mask_slots, num_classes] tensor, the
        class logits predicted by the transformer. It concats (num_classes - 1)
        non-void classes, including both thing classes and stuff classes, and the
        void class (the last channel). If the dataset class IDs do not follow this
        order, MaX-DeepLab loss functions will handle the mapping and thus the
        architecture still supports any dataset.
      void_prior_prob: A float, the desired probability (after softmax) of the
        void class at initialization. Defaults to 0.9 as in MaX-DeepLab.

    Returns:
      updated_transformer_class_logits: A [batch, num_mask_slots, num_classes]

    Raises:
      ValueError: If the rank of transformer_class_logits is not 3.
    """
    class_logits_shape = transformer_class_logits.get_shape().as_list()
    if len(class_logits_shape) != 3:
        raise ValueError('Input transformer_class_logits should have rank 3.')

    init_bias = [0.0] * class_logits_shape[-1]
    init_bias[-1] = math.log(
        (class_logits_shape[-1] - 1) * void_prior_prob / (1 - void_prior_prob))

    # Broadcasting the 1D init_bias to the 3D transformer_class_logits.
    return transformer_class_logits + tf.constant(init_bias, dtype=tf.float32)


def batch_norm_on_an_extra_axis(inputs, bn_layer):
    """Applies a batch norm layer on an extra axis.

    This batch norm will be used on the pixel space mask logits in MaX-DeepLab to
    avoid careful initialization of previous layers and careful scaling of the
    resulting outputs. In addition, applying batch norm on an extra axis does not
    introduce an extra gamma and beta for each mask slot. Instead, the current
    gamma and beta are shared for all mask slots and do not introduce biases on
    mask slots.

    Args:
      inputs: A [batch, height, width, num_mask_slots] tensor.
      bn_layer: A batch norm tf.keras.layers.Layer on the last axis.

    Returns:
      outputs: A [batch, height, width, num_mask_slots] tensor.
    """
    expanded_inputs = tf.expand_dims(inputs, axis=-1)
    outputs = bn_layer(expanded_inputs)
    return tf.squeeze(outputs, axis=-1)


class MaXDeepLab(tf.keras.layers.Layer):
    """A MaX-DeepLab head layer."""

    def __init__(self,
                 decoder_options,
                 max_deeplab_options,
                 ignore_label,
                 bn_layer=tf.keras.layers.BatchNormalization,
                 use_auxiliary_semantic_decoder=True,
                 use_auxiliary_semantic_head=True,
                 activation='relu'):
        """Initializes a MaX-DeepLab head.

        Args:
          decoder_options: Decoder options as defined in config_pb2.DecoderOptions.
          max_deeplab_options: Model options as defined in
            config_pb2.ModelOptions.MaXDeepLabOptions.
          ignore_label: An integer specifying the ignore label.
          bn_layer: An optional tf.keras.layers.Layer that computes the
            normalization (default: tf.keras.layers.BatchNormalization).
          use_auxiliary_semantic_decoder: A boolean, whether to use an axuliary
            semantic decoder to upsample the semantic features.
          use_auxiliary_semantic_head: A boolean, whether to use an auxiliary
            semantic head to generate semantic predictions.
          activation: A string, type of activation function to apply. Support
            'relu', 'swish' (or 'silu'), 'gelu', 'approximated_gelu', and 'elu'.
        """
        super(MaXDeepLab, self).__init__(name='MaXDeepLab')

        use_auxiliary_semantic_decoder = (
                use_auxiliary_semantic_decoder and use_auxiliary_semantic_head)
        if use_auxiliary_semantic_decoder:
            low_level_feature_keys = [
                item.feature_key for item in max_deeplab_options.auxiliary_low_level
            ]
            low_level_channels_project = [
                item.channels_project
                for item in max_deeplab_options.auxiliary_low_level
            ]

            self._auxiliary_semantic_decoder = (
                panoptic_deeplab.PanopticDeepLabSingleDecoder(
                    high_level_feature_name=decoder_options.feature_key,
                    low_level_feature_names=low_level_feature_keys,
                    low_level_channels_project=low_level_channels_project,
                    aspp_output_channels=decoder_options.aspp_channels,
                    decoder_output_channels=decoder_options.decoder_channels,
                    atrous_rates=decoder_options.atrous_rates,
                    name='auxiliary_semantic_decoder',
                    aspp_use_only_1x1_proj_conv=decoder_options
                    .aspp_use_only_1x1_proj_conv,
                    decoder_conv_type=decoder_options.decoder_conv_type,
                    bn_layer=bn_layer,
                    activation=activation))
        if use_auxiliary_semantic_head:
            self._auxiliary_semantic_head = (
                panoptic_deeplab.PanopticDeepLabSingleHead(
                    max_deeplab_options.auxiliary_semantic_head.head_channels,
                    max_deeplab_options.auxiliary_semantic_head.output_channels,
                    common.PRED_SEMANTIC_LOGITS_KEY,
                    name='auxiliary_semantic_head',
                    conv_type=max_deeplab_options.auxiliary_semantic_head
                    .head_conv_type,
                    bn_layer=bn_layer,
                    activation=activation))

        self._pixel_space_head = panoptic_deeplab.PanopticDeepLabSingleHead(
            max_deeplab_options.pixel_space_head.head_channels,
            max_deeplab_options.pixel_space_head.output_channels,
            _PIXEL_SPACE_FEATURE_KEY,
            name='pixel_space_head',
            conv_type=max_deeplab_options.pixel_space_head.head_conv_type,
            bn_layer=bn_layer,
            activation=activation)

        self._transformer_mask_head = convolutions.Conv1D(
            output_channels=max_deeplab_options.pixel_space_head.output_channels,
            name='transformer_mask_head',
            use_bias=False,
            # Use bn to avoid careful initialization.
            use_bn=True,
            bn_layer=bn_layer,
            bn_gamma_initializer='ones',
            activation=None,
            kernel_initializer='he_normal',
            kernel_size=1,
            padding='valid')
        # The transformer class head predicts non-void classes (i.e., thing classes
        # and stuff classes) and a void (i.e., ∅, no object) class.
        num_classes = _get_transformer_class_head_num_classes(
            max_deeplab_options.auxiliary_semantic_head.output_channels,
            ignore_label=ignore_label)
        self._transformer_class_head = convolutions.Conv1D(
            output_channels=num_classes,
            name='transformer_class_head',
            # Use conv bias rather than bn on this final class logit output.
            use_bias=True,
            use_bn=False,
            activation=None,
            # Follow common ImageNet class initlization with stddev 0.01.
            kernel_initializer=tf.keras.initializers.TruncatedNormal(stddev=0.01),
            kernel_size=1,
            padding='valid')

        self._pixel_space_feature_batch_norm = bn_layer(
            axis=-1, name='pixel_space_feature_batch_norm',
            gamma_initializer=tf.keras.initializers.Constant(1.0))
        # Use a batch norm to avoid care initialization of the mask outputs.
        self._pixel_space_mask_batch_norm = bn_layer(
            axis=-1, name='pixel_space_mask_batch_norm',
            # Initialize the pixel space mask with a low temperature.
            gamma_initializer=tf.keras.initializers.Constant(0.1))

        self._use_auxiliary_semantic_decoder = use_auxiliary_semantic_decoder
        self._use_auxiliary_semantic_head = use_auxiliary_semantic_head

    def reset_pooling_layer(self):
        """Resets the ASPP pooling layers to global average pooling."""
        if self._use_auxiliary_semantic_decoder:
            self._auxiliary_semantic_decoder.reset_pooling_layer()
        else:
            self._pool_size = (None, None)

    def set_pool_size(self, pool_size):
        """Sets the pooling size of the ASPP pooling layers.

        Args:
          pool_size: A tuple specifying the pooling size of the ASPP pooling layers.
        """
        if self._use_auxiliary_semantic_decoder:
            self._auxiliary_semantic_decoder.set_pool_size(pool_size)
        else:
            self._pool_size = pool_size

    def get_pool_size(self):
        if self._use_auxiliary_semantic_decoder:
            return self._auxiliary_semantic_decoder.get_pool_size()
        else:
            return self._pool_size

    @property
    def checkpoint_items(self):
        items = {
            common.CKPT_PIXEL_SPACE_HEAD:
                self._pixel_space_head,
            common.CKPT_TRANSFORMER_MASK_HEAD:
                self._transformer_mask_head,
            common.CKPT_TRANSFORMER_CLASS_HEAD:
                self._transformer_class_head,
            common.CKPT_PIXEL_SPACE_FEATURE_BATCH_NORM:
                self._pixel_space_feature_batch_norm,
            common.CKPT_PIXEL_SPACE_MASK_BATCH_NORM:
                self._pixel_space_mask_batch_norm,
        }
        if self._use_auxiliary_semantic_decoder:
            items.update({
                common.CKPT_SEMANTIC_DECODER:
                    self._auxiliary_semantic_decoder,
            })
        if self._use_auxiliary_semantic_head:
            items.update({
                common.CKPT_SEMANTIC_HEAD_WITHOUT_LAST_LAYER:
                    self._auxiliary_semantic_head.conv_block,
                common.CKPT_SEMANTIC_LAST_LAYER:
                    self._auxiliary_semantic_head.final_conv,
            })
        return items

    def call(self, features, training=False):
        """Performs a forward pass.

        Args:
          features: An input dict of tf.Tensor with shape [batch, height, width,
            channels] or [batch, length, channels]. Different keys should point to
            different features extracted by the encoder, e.g., low-level or
            high-level features.
          training: A boolean flag indicating whether training behavior should be
            used (default: False).

        Returns:
          A dictionary containing the auxiliary semantic segmentation logits, the
            pixel space normalized feature, the pixel space mask logits, and the
            mask transformer class logits.
        """
        results = {}
        semantic_features = features['feature_semantic']
        panoptic_features = features['feature_panoptic']
        transformer_class_feature = features['transformer_class_feature']
        transformer_mask_feature = features['transformer_mask_feature']

        if self._use_auxiliary_semantic_head:
            # Auxiliary semantic head.
            semantic_shape = semantic_features.get_shape().as_list()
            panoptic_shape = panoptic_features.get_shape().as_list()
            # MaX-DeepLab always predicts panoptic feature at high resolution (e.g.,
            # stride 4 or stride 2), but the auxiliary semantic feature could be at
            # low resolution (e.g., stride 16 or stride 32), in the absence of the
            # stacked decoder (L == 0). In this case, we use an auxiliary semantic
            # decoder on top of the semantic feature, in order to add the auxiliary
            # semantic loss.
            if (semantic_shape[1:3] != panoptic_shape[1:3] and
                    self._use_auxiliary_semantic_decoder):
                semantic_features = self._auxiliary_semantic_decoder(
                    features, training=training)
            auxiliary_semantic_results = self._auxiliary_semantic_head(
                semantic_features, training=training)
            results.update(auxiliary_semantic_results)

        # Pixel space head.
        pixel_space_feature = self._pixel_space_head(
            panoptic_features, training=training)[_PIXEL_SPACE_FEATURE_KEY]
        pixel_space_feature = self._pixel_space_feature_batch_norm(
            pixel_space_feature)
        pixel_space_normalized_feature = tf.math.l2_normalize(
            pixel_space_feature, axis=-1)
        results[common.PRED_PIXEL_SPACE_NORMALIZED_FEATURE_KEY] = (
            pixel_space_normalized_feature)

        # Transformer class head.
        transformer_class_logits = self._transformer_class_head(
            transformer_class_feature)
        # Bias towards the void class at initialization.
        transformer_class_logits = add_bias_towards_void(
            transformer_class_logits)
        results[common.PRED_TRANSFORMER_CLASS_LOGITS_KEY] = transformer_class_logits

        # Transformer mask kernel.
        transformer_mask_kernel = self._transformer_mask_head(
            transformer_mask_feature)

        # Convolutional mask head. The pixel space mask logits are the matrix
        # multiplication (or convolution) of the pixel space normalized feature and
        # the transformer mask kernel.
        pixel_space_mask_logits = tf.einsum(
            'bhwd,bid->bhwi',
            pixel_space_normalized_feature,
            transformer_mask_kernel)
        # The above multiplication constructs a second-order operation which is
        # sensitive to the feature scales and initializations. In order to avoid
        # careful initialization or scaling of the layers, we apply batch norms on
        # top of pixel_space_feature, transformer_mask_kernel, and the resulting
        # pixel_space_mask_logits.
        pixel_space_mask_logits = batch_norm_on_an_extra_axis(
            pixel_space_mask_logits, self._pixel_space_mask_batch_norm)
        results[common.PRED_PIXEL_SPACE_MASK_LOGITS_KEY] = (
            pixel_space_mask_logits)

        if common.PRED_AUXILIARY_OUTPUTS in features:
            results[common.PRED_AUXILIARY_OUTPUTS] = (
                features[common.PRED_AUXILIARY_OUTPUTS])

        return results





def demo_dataloader_dilate():
    import geowatch
    from geowatch.tasks.fusion.datamodules.kwcoco_dataset import KWCocoVideoDataset
    coco_dset = geowatch.coerce_kwcoco('vidshapes1', num_frames=10)
    self = KWCocoVideoDataset(coco_dset, time_dims=5,
                              window_dims=(512, 512),
                              balance_areas=True,
                              dist_weights=True,
                              normalize_perframe=False)
    self.disable_augmenter = True

    # Choose the target region we will sample
    target = self.new_sample_grid['targets'][self.new_sample_grid['positives_indexes'][0]]

    target['dist_weights'] = True
    item1 = self[target]

    target['dist_weights'] = False
    item2 = self[target]

    # Visualize the item
    canvas1 = self.draw_item(item1)
    canvas2 = self.draw_item(item2)

    import kwplot
    kwplot.autompl()
    kwplot.imshow(canvas1, pnum=(1, 2, 1), fnum=1, title='dist_weights=1')
    kwplot.imshow(canvas2, pnum=(1, 2, 2), fnum=1, title='dist_weights=0')
    kwplot.show_if_requested()



# coding=utf-8
# Copyright 2022 The Deeplab2 Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This file contains the attention used in MOAT [1].

The attention mechanism is based on CoAtNet [2].

[1] MOAT: Alternating Mobile Convolution and Attention
    Brings Strong Vision Models
    arXiv: 2210.01820.
      Chenglin Yang, Siyuan Qiao, Qihang Yu, Xiaoding Yuan,
      Yukun Zhu, Alan Yuille, Hartwig Adam, Liang-Chieh Chen.

[2] CoAtNet: Marrying Convolution and Attention for All Data Sizes
    NeurIPS 2021.
      Zihang Dai, Hanxiao Liu, Quoc V. Le, Mingxing Tan.
"""

import string
from typing import Optional

import numpy as np
import tensorflow as tf
from deeplab2.utils.hparam_configs import create_config_from_dict

# Global dict storing the computed lookup tensor to avoid repeated computation.
LOOKUP_TENSOR_CACHE = {}


def generate_lookup_tensor(length: int,
                           max_relative_position: Optional[int] = None,
                           dtype: tf.dtypes.DType = tf.float32) -> np.ndarray:
    """Generates one-hot lookup tensor to reindex embeddings along one dimension.

    Args:
      length: The length to reindex to.
      max_relative_position: The maximum relative position to consider.
        Relative position embeddings for distances above this threshold
        are zeroed out.
      dtype: The data type of the returned lookup tensor.

    Returns:
      lookup_tensor: A tensor with shape
      [length, length, relative_position_range]. The element
        satisfies lookup_tensor[n, m, v] = 1{m - n + max_relative_position = v},
        where n, m mean two positions while v the relative position selection.
    """
    if max_relative_position is None:
        max_relative_position = length - 1
    lookup_key = ('lookup_matrix', length, max_relative_position)
    relative_position_range = 2 * max_relative_position + 1
    lookup_tensor_shape = [length, length, relative_position_range]

    if (lookup_key not in LOOKUP_TENSOR_CACHE
            or LOOKUP_TENSOR_CACHE[lookup_key].shape.as_list() != lookup_tensor_shape
            or LOOKUP_TENSOR_CACHE[lookup_key].dtype != dtype):
        lookup_tensor = np.zeros(lookup_tensor_shape)
        for i in range(length):
            for x in range(length):
                v = x - i + max_relative_position
                if abs(x - i) > max_relative_position:
                    continue
                lookup_tensor[i, x, v] = 1
        LOOKUP_TENSOR_CACHE[lookup_key] = tf.constant(lookup_tensor, dtype)
    return LOOKUP_TENSOR_CACHE[lookup_key]


def reindex_2d_einsum_lookup(relative_position_tensor: tf.Tensor,
                             height: int,
                             width: int,
                             max_relative_height: Optional[int] = None,
                             max_relative_width: Optional[int] = None,
                             h_axis: int = 0) -> tf.Tensor:
    """Reindexes 2d relative position bias with 2 independent einsum lookups.

    Args:
      relative_position_tensor: A tensor of shape
        [..., relative_position_embedding_height,
        relative_position_embedding_width, ...].
      height: The height to reindex to.
      width: The width to reindex to.
      max_relative_height: Maximum relative height.
        Position embeddings corresponding to vertical distances larger
        than max_relative_height are zeroed out. None to disable.
      max_relative_width: Maximum relative width.
        Position embeddings corresponding to horizontal distances larger
        than max_relative_width are zeroed out. None to disable.
      h_axis: Axis corresponding to relative_position_embedding_height.
        Default to 0.

    Returns:
      reindexed_position_embedding: A Tensor of shape
        [..., height * width, height * width, ...]
    """
    height_lookup = generate_lookup_tensor(
        height, max_relative_position=max_relative_height,
        dtype=relative_position_tensor.dtype)
    width_lookup = generate_lookup_tensor(
        width, max_relative_position=max_relative_width,
        dtype=relative_position_tensor.dtype)

    non_spatial_rank = relative_position_tensor.shape.rank - 2
    non_spatial_expr = ''.join(chr(ord('n') + i) for i in range(non_spatial_rank))
    prefix = non_spatial_expr[:h_axis]
    suffix = non_spatial_expr[h_axis:]

    reindexed_tensor = tf.einsum(
        '{0}hw{1},ixh->{0}ixw{1}'.format(prefix, suffix),
        relative_position_tensor, height_lookup, name='height_lookup')
    reindexed_tensor = tf.einsum(
        '{0}ixw{1},jyw->{0}ijxy{1}'.format(prefix, suffix),
        reindexed_tensor, width_lookup, name='width_lookup')

    ret_shape = relative_position_tensor.shape.as_list()
    ret_shape[h_axis] = height * width
    ret_shape[h_axis + 1] = height * width
    reindexed_tensor = tf.reshape(reindexed_tensor, ret_shape)
    return reindexed_tensor


class TrailDense(tf.keras.layers.Layer):
    """A dense layer that projects features in multiple trailing axes.

    This layer projects features from multiple dimensions to multiple dimensions.
    The trailing axes with size n mean the last n dimensions. This layer avoids
    the extra uses of reshape operations.

    A einsum expression string is generated in this layer, examples:
      - For 4D tensors in conv, a common expression would be 'ABCD,DE->ABCE'.
      - For `q/k/v` head projection in multi-head attention with two output
        trailing dimensions, the expression is 'ABC,CDE->ABDE'
      - For `o` output projection in multi-head attention with
        input_begin_axis = -2, the expression is 'ABCD,CDE->ABE'
    """

    def __init__(self,
                 output_trailing_dimensions,
                 input_begin_axis=-1,
                 use_bias=True,
                 kernel_initializer=tf.random_normal_initializer(stddev=0.02),
                 bias_initializer=tf.zeros_initializer,
                 name='dense'):
        """Initializes TrailDense layer.

        Args:
          output_trailing_dimensions: A list of integers, multiple output
            dimensions in trailing axes. This avoids extra reshape
            operation that splits one single output dimension.
          input_begin_axis: A negative integer, the beginning axes of the input.
            This saves extra reshape operation to merge multiple input dimension.
          use_bias: A boolen, whether to use learnable bias in the layer.
          kernel_initializer: Initializer for the kernel weights matrix.
          bias_initializer: Initializer for the bias vector.
          name: A string, layer name.
        """

        super().__init__(name=name)
        self._output_trailing_dimensions = output_trailing_dimensions
        self._input_begin_axis = input_begin_axis
        self._use_bias = use_bias
        self._kernel_initializer = kernel_initializer
        self._bias_initializer = bias_initializer

    def build(self, input_shape):
        """Creates variables and einsum expression based on input shape."""

        weight_shape = (input_shape[self._input_begin_axis:] +
                        self._output_trailing_dimensions)
        self.weight = self.add_weight(
            name='weight',
            shape=weight_shape,
            initializer=self._kernel_initializer,
            trainable=True)
        if self._use_bias:
            self.bias = self.add_weight(
                name='bias',
                shape=self._output_trailing_dimensions,
                initializer=self._bias_initializer,
                trainable=True)

        # Create einsum expression.
        input_rank = input_shape.rank
        shared_size = self._input_begin_axis % input_rank
        i_only_size = input_rank - shared_size
        o_only_size = len(self._output_trailing_dimensions)

        if input_rank + o_only_size >= len(string.ascii_uppercase):
            raise ValueError('Cannot use einsum as input rank + output rank > 26.')
        einsum_str = string.ascii_uppercase[:input_rank + o_only_size]

        offset = 0
        shared_str = einsum_str[offset:offset + shared_size]
        offset += shared_size
        i_only_str = einsum_str[offset:offset + i_only_size]
        offset += i_only_size
        o_only_str = einsum_str[offset:offset + o_only_size]

        input_str = '{}{}'.format(shared_str, i_only_str)
        output_str = '{}{}'.format(shared_str, o_only_str)
        weight_str = '{}{}'.format(i_only_str, o_only_str)
        self.einsum_expr = '{},{}->{}'.format(input_str, weight_str, output_str)

    def call(self, inputs):
        output = tf.einsum(self.einsum_expr, inputs, self.weight)
        if self._use_bias:
            output += self.bias
        return output


class Attention(tf.keras.layers.Layer):
    """Implementation of Attention.

    This layer performs global self-attention [1] on the input. The input shape
    is [batch_size, height, width, channels] and the output shape is
    [batch_size, height * width, channels].

    If one would like to extend the global self-attention to the local window
    attention [2], they could reshape the input to
    [batch_size * num_window, pixel_num_per_window, channels] followed by
    applying this class.

    [1] Attention Is All You Need.
      NeurIPS 2017.
        Ashish Vaswani, Noam Shazeer, Niki Parmar, Jakob Uszkoreit, Llion Jones,
        Aidan N. Gomez, Lukasz Kaiser, Illia Polosukhin.

    [2] Swin Transformer: Hierarchical Vision Transformer using Shifted Windows.
      ICCV 2021.
        Ze Liu, Yutong Lin, Yue Cao, Han Hu, Yixuan Wei, Zheng Zhang,
        Stephen Lin, Baining Guo
    """

    def _retrieve_config(self, config):
        """Retrieves the config of Attention.

        Args:
          config: A dictionary containing the following keys.
            -hidden_size: An integer, output channels.
            -head_size: An integer, the head size of attention.
            -relative_position_embedding_type: A string, type of relative position
              embedding in Attention. Only '2d_multi_head' is supported now.
              If None, no relative position embedding will be used.
            -scale_ratio: A float or a list of floats with length 2, scaling factors
              for the position embedding in height and width dimensions. For
              example, [14/14, 16/14] means the position embedding is created for
              window 14 x 14, but will be interpolated to 14 x 16.
            -kernel_initializer: Initializer for the kernel weights matrix.
            -bias_initializer: Initializer for the bias vector.
            -name: A string, layer name.

        Returns:
          A Config class: hparams_config.Config.
        """

        required_keys = ['hidden_size', 'head_size']
        optional_keys = {
            'relative_position_embedding_type': None,
            'scale_ratio': None,
            'kernel_initializer': tf.random_normal_initializer(stddev=0.02),
            'bias_initializer': tf.zeros_initializer,
            'name': 'attention',
        }
        config = create_config_from_dict(config, required_keys, optional_keys)
        return config

    def __init__(self, **config):
        self._config = self._retrieve_config(config)
        super().__init__(name=self._config.name)
        self._config.num_heads = self._config.hidden_size // self._config.head_size

        self._q_proj = TrailDense(
            output_trailing_dimensions=[self._config.num_heads,
                                        self._config.head_size],
            kernel_initializer=self._config.kernel_initializer,
            bias_initializer=self._config.bias_initializer,
            name='q')
        self._k_proj = TrailDense(
            output_trailing_dimensions=[self._config.num_heads,
                                        self._config.head_size],
            kernel_initializer=self._config.kernel_initializer,
            bias_initializer=self._config.bias_initializer,
            name='k')
        self._v_proj = TrailDense(
            output_trailing_dimensions=[self._config.num_heads,
                                        self._config.head_size],
            kernel_initializer=self._config.kernel_initializer,
            bias_initializer=self._config.bias_initializer,
            name='v')
        self._o_proj = TrailDense(
            output_trailing_dimensions=[self._config.hidden_size],
            input_begin_axis=-2,
            kernel_initializer=self._config.kernel_initializer,
            bias_initializer=self._config.bias_initializer,
            name='o')

        self._q_scale = self._config.head_size ** -0.5

    def build(self, input_shape):
        if self._config.relative_position_embedding_type == '2d_multi_head':
            if input_shape.rank != 4:
                raise ValueError(
                    'The input shape should be [batch_size, height, width, channels]')
            input_shape_list = input_shape.as_list()
            height, width = input_shape_list[-3], input_shape_list[-2]
            if self._config.scale_ratio is not None:
                if (isinstance(self._config.scale_ratio, list) and
                        len(self._config.scale_ratio) == 2):
                    height_scale, width_scale = self._config.scale_ratio
                elif isinstance(self._config.scale_ratio, float):
                    height_scale = self._config.scale_ratio
                    width_scale = self._config.scale_ratio
                else:
                    raise ValueError(
                        'scale ratio should be float or list of floats with length 2')

                relative_position_embedding_height = (
                        2 * round(height / height_scale) - 1)
                relative_position_embedding_width = (
                        2 * round(width / width_scale) - 1)
            else:
                relative_position_embedding_height = 2 * height - 1
                relative_position_embedding_width = 2 * width - 1
            relative_position_embedding_height_axis = 1
            relative_position_embedding_shape = [
                self._config.num_heads,
                relative_position_embedding_height,
                relative_position_embedding_width]
            self.relative_position_embedding = self.add_weight(
                'relative_position_embedding',
                relative_position_embedding_shape,
                initializer=self._config.kernel_initializer,
                trainable=True)
            if self._config.scale_ratio is not None:
                relative_position_embedding = tf.expand_dims(
                    self.relative_position_embedding, axis=-1)
                relative_position_embedding = tf.cast(
                    tf.image.resize(relative_position_embedding,
                                    [2 * height - 1, 2 * width - 1]),
                    self.compute_dtype)
                relative_position_embedding = tf.squeeze(relative_position_embedding,
                                                         axis=-1)
            else:
                relative_position_embedding = tf.cast(self.relative_position_embedding,
                                                      self.compute_dtype)

            self.reindexed_position_embedding = reindex_2d_einsum_lookup(
                relative_position_embedding, height, width, height - 1, width - 1,
                h_axis=relative_position_embedding_height_axis)
        elif self._config.relative_position_embedding_type is None:
            self.reindexed_position_embedding = None

    def call(self, query, training):
        _, h, w, channels = query.shape.as_list()
        query = tf.reshape(query, [-1, h * w, channels])

        q_heads = self._q_proj(query)
        k_heads = self._k_proj(query)
        v_heads = self._v_proj(query)
        q_heads *= self._q_scale

        attention_logits = tf.einsum('BSNK, BTNK -> BNST', q_heads, k_heads)

        if self.reindexed_position_embedding is not None:
            attention_logits += self.reindexed_position_embedding

        attention_probs = tf.cast(
            tf.nn.softmax(tf.cast(attention_logits, tf.float32), axis=-1),
            attention_logits.dtype)

        attention_out = tf.einsum('BNST, BTNK -> BSNK', attention_probs, v_heads)
        output = self._o_proj(attention_out)
        return output



"""
Can we make scriptconfig handle what we really need?
"""
#!/usr/bin/env python3
import scriptconfig as scfg
from geowatch.tasks.fusion._lightning_components import SmartTrainer
from geowatch.tasks.fusion.methods import MultimodalTransformer  # NOQA
from geowatch.tasks.fusion.datamodules.kwcoco_datamodule import KWCocoVideoDataModule
# import ubelt as ub


class PocAltFitCLI(scfg.DataConfig):
    """
    import sys, ubelt
    sys.path.append(ubelt.expandpath('~/code/geowatch'))
    from geowatch.tasks.fusion._poc_alt_fit import *  # NOQA
    """
    trainer = scfg.Value(None, help='A YAML dictionary of updates to the defaults.')
    data = scfg.Value(None, help='A YAML dictionary of updates to the defaults.')
    model = scfg.Value(None, help='A YAML dictionary of updates to the defaults.')
    optimizer = scfg.Value(None, help='A YAML dictionary of updates to the defaults.')
    lr_scheduler = scfg.Value(None, help='A YAML dictionary of updates to the defaults.')
    torch_globals = scfg.Value(None, help='A YAML dictionary of updates to the defaults.')
    initializer = scfg.Value(None, help='A YAML dictionary of updates to the defaults.')


def main():
    """

    config = ub.codeblock(
        '''
        data:
            class_path: KWCocoVideoDataModule
            init_args:
                train_dataset: special:vidshapes8-frames9-speed0.5
                window_dims: 64
                num_workers: 4
                batch_size: 4
                normalize_inputs:
                    input_stats:
                        - sensor: '*'
                          channels: r|g|b
                          video: video1
                          mean: [87.572401, 87.572402, 87.572403]
                          std: [99.449997, 99.449998, 99.449999]
        model:
            class_path: MultimodalTransformer
        optimizer:
            class_path: torch.optim.Adam
        trainer:
            accelerator: gpu
            devices: 1
            default_root_dir: ./demo_train
        ''')
    """
    import ubelt as ub

    defaults = dict(
        trainer={'class_path': SmartTrainer, 'init_args': {}},
        model={'class_path': MultimodalTransformer, 'init_args': {}},
        data={'class_path': KWCocoVideoDataModule, 'init_args': {}},
        optimizer={'class_path': 'torch.optim.AdamW', 'init_args': {}},
        lr_scheduler={'class_path': 'torch.optim.lr_scheduler.ConstantLR', 'init_args': {}},
    )

    __path_defaults__ = {}

    for key, data in defaults.items():
        new_data = coerce_instance_defaults(key, data)

        for param in new_data:
            path = f'{key}.{param.name}'
            __path_defaults__[path] = scfg.Value(param.default, help=None if param.doc is None else ub.paragraph(param.doc))

    print(f'__path_defaults__ = {ub.urepr(__path_defaults__, nl=1)}')


def coerce_instance_defaults(key, data):
    import torch  # NOQA
    import kwutil
    from geowatch.tasks.fusion import _jsonargparse_introspect
    data = kwutil.Yaml.coerce(data)
    if isinstance(data, dict):
        if 'class_path' in data:
            cls = data['class_path']
            if isinstance(cls, str):
                cls = eval(cls, globals(), locals())

            default_params = _jsonargparse_introspect.get_signature_parameters(cls)
            return default_params

            if 'init_args' in data:
                init_kw = data['init_args']
            else:
                init_kw = {}
        else:
            raise NotImplementedError
        new_data = {'cls': cls, 'init_kw': init_kw}
        return
    else:
        raise NotImplementedError

if __name__ == '__main__':
    """

    CommandLine:
        python ~/code/geowatch/geowatch/tasks/fusion/_poc_alt_fit.py
        python -m geowatch.tasks.fusion._poc_alt_fit
    """
    main()



# coding=utf-8
# Copyright 2022 The Deeplab2 Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This file contains functions to build encoder and decoder."""
from typing import Callable, List, Optional

import tensorflow as tf
from deeplab2 import config_pb2
from deeplab2.model.decoder import deeplabv3
from deeplab2.model.decoder import deeplabv3plus
from deeplab2.model.decoder import max_deeplab
from deeplab2.model.decoder import motion_deeplab_decoder
from deeplab2.model.decoder import panoptic_deeplab
from deeplab2.model.decoder import vip_deeplab_decoder
from deeplab2.model.encoder import axial_resnet_instances
from deeplab2.model.encoder import mobilenet
from deeplab2.model.pixel_decoder import kmax as kmax_pixel_decoder
from deeplab2.model.pixel_encoder import axial_resnet
from deeplab2.model.pixel_encoder import convnext
from deeplab2.model.pixel_encoder import moat
from deeplab2.model.transformer_decoder import kmax as kmax_trasnformer_decoder


def create_encoder(backbone_options: config_pb2.ModelOptions.BackboneOptions,
                   bn_layer: tf.keras.layers.Layer,
                   conv_kernel_weight_decay: float = 0.0) -> tf.keras.Model:
    """Creates an encoder.

    Args:
      backbone_options: A proto config of type
        config_pb2.ModelOptions.BackboneOptions.
      bn_layer: A tf.keras.layers.Layer that computes the normalization.
      conv_kernel_weight_decay: A float, the weight decay for convolution kernels.

    Returns:
      An instance of tf.keras.Model containing the encoder.

    Raises:
      ValueError: An error occurs when the specified encoder meta architecture is
        not supported.
    """
    if ('resnet' in backbone_options.name or
            'swidernet' in backbone_options.name or
            'axial_deeplab' in backbone_options.name or
            'max_deeplab' in backbone_options.name):
        return create_resnet_encoder(
            backbone_options,
            bn_layer=bn_layer,
            conv_kernel_weight_decay=conv_kernel_weight_decay)
    elif 'mobilenet' in backbone_options.name:
        return create_mobilenet_encoder(
            backbone_options,
            bn_layer=bn_layer,
            conv_kernel_weight_decay=conv_kernel_weight_decay)
    raise ValueError('The specified encoder %s is not a valid encoder.' %
                     backbone_options.name)


def create_mobilenet_encoder(
        backbone_options: config_pb2.ModelOptions.BackboneOptions,
        bn_layer: tf.keras.layers.Layer,
        conv_kernel_weight_decay: float = 0.0) -> tf.keras.Model:
    """Creates a MobileNet encoder specified by name.

    Args:
      backbone_options: A proto config of type
        config_pb2.ModelOptions.BackboneOptions.
      bn_layer: A tf.keras.layers.Layer that computes the normalization.
      conv_kernel_weight_decay: A float, the weight decay for convolution kernels.

    Returns:
      An instance of tf.keras.Model containing the MobileNet encoder.
    """
    if backbone_options.name.lower() == 'mobilenet_v3_large':
        backbone = mobilenet.MobileNetV3Large
    elif backbone_options.name.lower() == 'mobilenet_v3_small':
        backbone = mobilenet.MobileNetV3Small
    else:
        raise ValueError('The specified encoder %s is not a valid encoder.' %
                         backbone_options.name)
    assert backbone_options.use_squeeze_and_excite
    assert backbone_options.drop_path_keep_prob == 1
    assert backbone_options.use_sac_beyond_stride == -1
    assert backbone_options.backbone_layer_multiplier == 1
    return backbone(
        output_stride=backbone_options.output_stride,
        width_multiplier=backbone_options.backbone_width_multiplier,
        bn_layer=bn_layer,
        conv_kernel_weight_decay=conv_kernel_weight_decay)


def create_resnet_encoder(
        backbone_options: config_pb2.ModelOptions.BackboneOptions,
        bn_layer: tf.keras.layers.Layer,
        conv_kernel_weight_decay: float = 0.0) -> tf.keras.Model:
    """Creates a ResNet encoder specified by name.

    Args:
      backbone_options: A proto config of type
        config_pb2.ModelOptions.BackboneOptions.
      bn_layer: A tf.keras.layers.Layer that computes the normalization.
      conv_kernel_weight_decay: A float, the weight decay for convolution kernels.

    Returns:
      An instance of tf.keras.Model containing the ResNet encoder.
    """
    return axial_resnet_instances.get_model(
        backbone_options.name,
        output_stride=backbone_options.output_stride,
        stem_width_multiplier=backbone_options.stem_width_multiplier,
        width_multiplier=backbone_options.backbone_width_multiplier,
        backbone_layer_multiplier=backbone_options.backbone_layer_multiplier,
        block_group_config={
            'use_squeeze_and_excite': backbone_options.use_squeeze_and_excite,
            'drop_path_keep_prob': backbone_options.drop_path_keep_prob,
            'drop_path_schedule': backbone_options.drop_path_schedule,
            'use_sac_beyond_stride': backbone_options.use_sac_beyond_stride},
        bn_layer=bn_layer,
        conv_kernel_weight_decay=conv_kernel_weight_decay)


def create_decoder(model_options: config_pb2.ModelOptions,
                   bn_layer: tf.keras.layers.Layer,
                   ignore_label: int,
                   use_auxiliary_semantic_head: bool = True) -> tf.keras.Model:
    """Creates a DeepLab decoder.

    Args:
      model_options: A proto config of type config_pb2.ModelOptions.
      bn_layer: A tf.keras.layers.Layer that computes the normalization.
      ignore_label: An integer specifying the ignore label.
      use_auxiliary_semantic_head: A boolean, whether to use an auxiliary semantic
        head to generate semantic predictions.

    Returns:
      An instance of tf.keras.layers.Layer containing the decoder.

    Raises:
      ValueError: An error occurs when the specified meta architecture is not
        supported.
    """
    meta_architecture = model_options.WhichOneof('meta_architecture')
    if meta_architecture == 'deeplab_v3':
        return deeplabv3.DeepLabV3(
            model_options.decoder, model_options.deeplab_v3, bn_layer=bn_layer)
    elif meta_architecture == 'deeplab_v3_plus':
        return deeplabv3plus.DeepLabV3Plus(
            model_options.decoder, model_options.deeplab_v3_plus, bn_layer=bn_layer)
    elif meta_architecture == 'panoptic_deeplab':
        return panoptic_deeplab.PanopticDeepLab(
            model_options.decoder,
            model_options.panoptic_deeplab,
            bn_layer=bn_layer)
    elif meta_architecture == 'motion_deeplab':
        return motion_deeplab_decoder.MotionDeepLabDecoder(
            model_options.decoder,
            model_options.motion_deeplab,
            bn_layer=bn_layer)
    elif meta_architecture == 'vip_deeplab':
        return vip_deeplab_decoder.ViPDeepLabDecoder(
            model_options.decoder,
            model_options.vip_deeplab,
            bn_layer=bn_layer)
    elif meta_architecture == 'max_deeplab':
        return max_deeplab.MaXDeepLab(
            model_options.decoder,
            model_options.max_deeplab,
            ignore_label=ignore_label,
            bn_layer=bn_layer,
            use_auxiliary_semantic_head=use_auxiliary_semantic_head)
    raise ValueError('The specified meta architecture %s is not implemented.' %
                     meta_architecture)


def create_kmax_meta_pixel_encoder(
        pixel_encoder_name: str,
        drop_path_keep_prob: float,
        input_shape: List[int],
        norm_layer: Optional[Callable[[], tf.keras.layers.Layer]],
        pretrained_weights_path: Optional[str] = None,
) -> tf.keras.Model:
    """Creates a pixel encoder in kMaX meta architecture.

    Args:
      pixel_encoder_name: A string specifying the pixel encoder name.
      drop_path_keep_prob: A float specifying the keep probability of drop path.
      input_shape: A list of integer specifying the expected input shape.
      norm_layer: A tf.keras.layers.Layer that computes the normalization.
      pretrained_weights_path: A string specifying the path of pretrained weights
        for the backbone (i.e., pixel encoder).

    Returns:
      An instance of tf.keras.Model containing the pixel encoder.

    Raises:
      ValueError: An error occurs when the pixel_encoder_name does not contain
        'convnext' or 'resnet'.
    """
    if 'convnext' in pixel_encoder_name.lower():
        return convnext.get_model(
            pixel_encoder_name,
            input_shape=input_shape,
            pretrained_weights_path=pretrained_weights_path,
            drop_path_keep_prob=drop_path_keep_prob,
            zero_padding_for_downstream=True)
    elif 'moat' in pixel_encoder_name.lower():
        return moat.get_model(
            name=pixel_encoder_name,
            input_shape=input_shape,
            survival_rate=drop_path_keep_prob,
            pretrained_weights_path=pretrained_weights_path,
        )
    elif 'resnet' in pixel_encoder_name.lower():
        return axial_resnet.get_model(
            pixel_encoder_name,
            input_shape=input_shape,
            bn_layer=norm_layer,
            drop_path_keep_prob=drop_path_keep_prob,
        )
    else:
        raise ValueError('Unsupoorted pixel_encoder_name!')


def create_kmax_meta_pixel_decoder(
        norm_layer: Optional[Callable[[], tf.keras.layers.Layer]],
        high_resolution_output_stride: int = 4
) -> tf.keras.Model:
    """Creates a pixel decoder in kMaX meta architecture.

    Args:
      norm_layer: A tf.keras.layers.Layer that computes the normalization.
      high_resolution_output_stride: An integer speicying the highest resolution
        of pixel decoder output feature maps.

    Returns:
      An instance of tf.keras.Model containing the pixel decoder.

    Raises:
      ValueError: An error occurs when the high_resolution_output_stride is not
        4 or 2.
    """
    # high_resolution_output_stride = 4 is used for COCO dataset.
    if high_resolution_output_stride == 4:
        dims = (512, 256, 128, 64)
        num_blocks = (1, 5, 1, 1)
        block_type = ('axial', 'axial', 'bottleneck', 'bottleneck')
    # high_resolution_output_stride = 2 is used for Cityscapes dataset.
    elif high_resolution_output_stride == 2:
        dims = (512, 256, 128, 64, 32)
        num_blocks = (1, 5, 1, 1, 1)
        block_type = ('axial', 'axial', 'bottleneck', 'bottleneck', 'bottleneck')
    else:
        raise ValueError('Unsupported high_resolution_output_stride for building'
                         ' kmax_meta_pixel_decoder!')

    return kmax_pixel_decoder.KMaXPixelDecoder(
        name='kmax_pixel_decoder',
        norm_layer=norm_layer,
        dims=dims,
        num_blocks=num_blocks,
        block_type=block_type)


def create_kmax_meta_transformer_decoder(
        norm_layer: Optional[Callable[[], tf.keras.layers.Layer]],
        num_mask_slots: int,
        transformer_decoder_drop_path_keep_prob: float,
        auxiliary_predictor_func: Optional[Callable[[], tf.keras.Model]] = None
) -> tf.keras.Model:
    """Creates a transformer decoder in kMaX meta architecture.

    Args:
      norm_layer: A tf.keras.layers.Layer that computes the normalization.
      num_mask_slots: An integer, the number of mask slots that will be used.
      transformer_decoder_drop_path_keep_prob: A float, the drop-path keep prob
        for transformer decoder.
      auxiliary_predictor_func: A callable function that returns an initialization
        of auxiliary predictor.

    Returns:
      An instance of tf.keras.Model containing the transformer decoder.
    """
    return kmax_trasnformer_decoder.KMaXTransformerDecoder(
        name='kmax_transformer_decoder',
        auxiliary_predictor_func=auxiliary_predictor_func,
        norm_layer=norm_layer,
        num_mask_slots=num_mask_slots,
        transformer_decoder_drop_path_keep_prob=(
            transformer_decoder_drop_path_keep_prob),
        num_blocks=(2, 2, 2))



# coding=utf-8
# Copyright 2022 The Deeplab2 Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This file contains code to build a Panoptic-DeepLab decoder.

Reference:
  - [Panoptic-DeepLab: A Simple, Strong, and Fast Baseline for Bottom-Up
      Panoptic Segmentation](https://arxiv.org/pdf/1911.10194)
"""
import tensorflow as tf
from absl import logging
from deeplab2 import common
from deeplab2.model import utils
from deeplab2.model.decoder import aspp
from deeplab2.model.layers import convolutions

layers = tf.keras.layers


class PanopticDeepLabSingleDecoder(layers.Layer):
    """A single Panoptic-DeepLab decoder layer.

    This layer takes low- and high-level features as input and uses an ASPP
    followed by a fusion block to decode features for a single task, e.g.,
    semantic segmentation or instance segmentation.
    """

    def __init__(self,
                 high_level_feature_name,
                 low_level_feature_names,
                 low_level_channels_project,
                 aspp_output_channels,
                 decoder_output_channels,
                 atrous_rates,
                 name,
                 aspp_use_only_1x1_proj_conv=False,
                 decoder_conv_type='depthwise_separable_conv',
                 bn_layer=tf.keras.layers.BatchNormalization,
                 activation='relu'):
        """Initializes a single Panoptic-DeepLab decoder of layers.Layer.

        Args:
          high_level_feature_name: A string specifying the name of the high-level
            feature coming from an encoder.
          low_level_feature_names: A list of strings specifying the name of the
            low-level features coming from an encoder. An order from highest to
            lower level is expected, e.g. ['res3', 'res2'].
          low_level_channels_project: A list of integer specifying the number of
            filters used for processing each low_level features.
          aspp_output_channels: An integer specifying the number of filters in the
            ASPP convolution layers.
          decoder_output_channels: An integer specifying the number of filters in
            the decoder convolution layers.
          atrous_rates: A list of three integers specifying the atrous rate for the
            ASPP layers.
          name: A string specifying the name of the layer.
          aspp_use_only_1x1_proj_conv: Boolean, specifying if the ASPP five branches
            are turned off or not. If True, the ASPP module is degenerated to one
            1x1 convolution, projecting the input channels to `output_channels`.
          decoder_conv_type: String, specifying decoder convolution type. Support
            'depthwise_separable_conv' and 'standard_conv'.
          bn_layer: An optional tf.keras.layers.Layer that computes the
            normalization (default: tf.keras.layers.BatchNormalization).
          activation: A string, type of activation function to apply. Support
            'relu', 'swish' (or 'silu'), 'gelu', 'approximated_gelu', and 'elu'.

        Raises:
          ValueError: An error occurs when the length of low_level_feature_names
            differs from the length of low_level_channels_project.
        """
        super(PanopticDeepLabSingleDecoder, self).__init__(name=name)
        self._channel_axis = 3

        self._aspp = aspp.ASPP(
            aspp_output_channels,
            atrous_rates,
            aspp_use_only_1x1_proj_conv=aspp_use_only_1x1_proj_conv,
            name='aspp',
            bn_layer=bn_layer,
            activation=activation)
        self._high_level_feature_name = high_level_feature_name

        if len(low_level_feature_names) != len(low_level_channels_project):
            raise ValueError('The Panoptic-DeepLab decoder requires the same number '
                             'of low-level features as the number of low-level '
                             'projection channels. But got %d and %d.'
                             % (len(low_level_feature_names),
                                len(low_level_channels_project)))

        self._low_level_feature_names = low_level_feature_names

        for i, channels_project in enumerate(low_level_channels_project):
            # Check if channel sizes increases and issue a warning.
            if i > 0 and low_level_channels_project[i - 1] < channels_project:
                logging.warning(
                    'The low level projection channels usually do not '
                    'increase for features with higher spatial resolution. '
                    'Please make sure, this behavior is intended.')
            current_low_level_conv_name, current_fusion_conv_name = (
                utils.get_low_level_conv_fusion_conv_current_names(i))
            utils.safe_setattr(
                self, current_low_level_conv_name, convolutions.Conv2DSame(
                    channels_project,
                    kernel_size=1,
                    name=utils.get_layer_name(current_low_level_conv_name),
                    use_bias=False,
                    use_bn=True,
                    bn_layer=bn_layer,
                    activation=activation))

            utils.safe_setattr(
                self, current_fusion_conv_name, convolutions.StackedConv2DSame(
                    conv_type=decoder_conv_type,
                    num_layers=1,
                    output_channels=decoder_output_channels,
                    kernel_size=5,
                    name=utils.get_layer_name(current_fusion_conv_name),
                    use_bias=False,
                    use_bn=True,
                    bn_layer=bn_layer,
                    activation=activation))

    def call(self, features, training=False):
        """Performs a forward pass.

        Args:
          features: An input dict of tf.Tensor with shape [batch, height, width,
            channels]. Different keys should point to different features extracted
            by the encoder, e.g. low-level or high-level features.
          training: A boolean flag indicating whether training behavior should be
            used (default: False).

        Returns:
          Refined features as instance of tf.Tensor.
        """

        high_level_features = features[self._high_level_feature_name]
        combined_features = self._aspp(high_level_features, training=training)

        # Fuse low-level features with high-level features.
        for i in range(len(self._low_level_feature_names)):
            current_low_level_conv_name, current_fusion_conv_name = (
                utils.get_low_level_conv_fusion_conv_current_names(i))
            # Iterate from the highest level of the low level features to the lowest
            # level, i.e. take the features with the smallest spatial size first.
            low_level_features = features[self._low_level_feature_names[i]]
            low_level_features = getattr(self, current_low_level_conv_name)(
                low_level_features, training=training)

            target_h = tf.shape(low_level_features)[1]
            target_w = tf.shape(low_level_features)[2]
            source_h = tf.shape(combined_features)[1]
            source_w = tf.shape(combined_features)[2]

            tf.assert_less(
                source_h - 1,
                target_h,
                message='Features are down-sampled during decoder.')
            tf.assert_less(
                source_w - 1,
                target_w,
                message='Features are down-sampled during decoder.')

            combined_features = utils.resize_align_corners(combined_features,
                                                           [target_h, target_w])

            combined_features = tf.concat([combined_features, low_level_features],
                                          self._channel_axis)
            combined_features = getattr(self, current_fusion_conv_name)(
                combined_features, training=training)

        return combined_features

    def reset_pooling_layer(self):
        """Resets the ASPP pooling layer to global average pooling."""
        self._aspp.reset_pooling_layer()

    def set_pool_size(self, pool_size):
        """Sets the pooling size of the ASPP pooling layer.

        Args:
          pool_size: A tuple specifying the pooling size of the ASPP pooling layer.
        """
        self._aspp.set_pool_size(pool_size)

    def get_pool_size(self):
        return self._aspp.get_pool_size()


class PanopticDeepLabSingleHead(layers.Layer):
    """A single PanopticDeepLab head layer.

    This layer takes in the enriched features from a decoder and adds two
    convolutions on top.
    """

    def __init__(self,
                 intermediate_channels,
                 output_channels,
                 pred_key,
                 name,
                 conv_type='depthwise_separable_conv',
                 bn_layer=tf.keras.layers.BatchNormalization,
                 activation='relu'):
        """Initializes a single PanopticDeepLab head.

        Args:
          intermediate_channels: An integer specifying the number of filters of the
            first 5x5 convolution.
          output_channels: An integer specifying the number of filters of the second
            1x1 convolution.
          pred_key: A string specifying the key of the output dictionary.
          name: A string specifying the name of this head.
          conv_type: String, specifying head convolution type. Support
            'depthwise_separable_conv' and 'standard_conv'.
          bn_layer: An optional tf.keras.layers.Layer that computes the
            normalization (default: tf.keras.layers.BatchNormalization).
          activation: A string, type of activation function to apply. Support
            'relu', 'swish' (or 'silu'), 'gelu', 'approximated_gelu', and 'elu'.
        """
        super(PanopticDeepLabSingleHead, self).__init__(name=name)
        self._pred_key = pred_key

        self.conv_block = convolutions.StackedConv2DSame(
            conv_type=conv_type,
            num_layers=1,
            output_channels=intermediate_channels,
            kernel_size=5,
            name='conv_block',
            use_bias=False,
            use_bn=True,
            bn_layer=bn_layer,
            activation=activation)
        self.final_conv = layers.Conv2D(
            output_channels,
            kernel_size=1,
            name='final_conv',
            kernel_initializer=tf.keras.initializers.TruncatedNormal(stddev=0.01))

    def call(self, features, training=False):
        """Performs a forward pass.

        Args:
          features: A tf.Tensor with shape [batch, height, width, channels].
          training: A boolean flag indicating whether training behavior should be
            used (default: False).

        Returns:
          The dictionary containing the predictions under the specified key.
        """
        x = self.conv_block(features, training=training)
        return {self._pred_key: self.final_conv(x)}


class PanopticDeepLab(layers.Layer):
    """A Panoptic-DeepLab decoder layer.

    This layer takes low- and high-level features as input and uses a dual-ASPP
    and dual-decoder structure to aggregate features for semantic and instance
    segmentation. On top of the decoders, three heads are used to predict semantic
    segmentation, instance center probabilities, and instance center regression
    per pixel.
    """

    def __init__(self,
                 decoder_options,
                 panoptic_deeplab_options,
                 bn_layer=tf.keras.layers.BatchNormalization,
                 activation='relu'):
        """Initializes a Panoptic-DeepLab decoder.

        Args:
          decoder_options: Decoder options as defined in config_pb2.DecoderOptions.
          panoptic_deeplab_options: Model options as defined in
            config_pb2.ModelOptions.PanopticDeeplabOptions.
          bn_layer: An optional tf.keras.layers.Layer that computes the
            normalization (default: tf.keras.layers.BatchNormalization).
          activation: A string, type of activation function to apply. Support
            'relu', 'swish' (or 'silu'), 'gelu', 'approximated_gelu', and 'elu'.
        """
        super(PanopticDeepLab, self).__init__(name='PanopticDeepLab')

        low_level_feature_keys = [
            item.feature_key for item in panoptic_deeplab_options.low_level
        ]
        low_level_channels_project = [
            item.channels_project for item in panoptic_deeplab_options.low_level
        ]

        self._semantic_decoder = PanopticDeepLabSingleDecoder(
            high_level_feature_name=decoder_options.feature_key,
            low_level_feature_names=low_level_feature_keys,
            low_level_channels_project=low_level_channels_project,
            aspp_output_channels=decoder_options.aspp_channels,
            decoder_output_channels=decoder_options.decoder_channels,
            atrous_rates=decoder_options.atrous_rates,
            name='semantic_decoder',
            aspp_use_only_1x1_proj_conv=decoder_options.aspp_use_only_1x1_proj_conv,
            decoder_conv_type=decoder_options.decoder_conv_type,
            bn_layer=bn_layer,
            activation=activation)
        self._semantic_head = PanopticDeepLabSingleHead(
            panoptic_deeplab_options.semantic_head.head_channels,
            panoptic_deeplab_options.semantic_head.output_channels,
            common.PRED_SEMANTIC_LOGITS_KEY,
            name='semantic_head',
            conv_type=panoptic_deeplab_options.semantic_head.head_conv_type,
            bn_layer=bn_layer,
            activation=activation)

        self._instance_decoder = None
        self._instance_center_head = None
        self._instance_regression_head = None

        if panoptic_deeplab_options.instance.enable:
            if panoptic_deeplab_options.instance.low_level_override:
                low_level_options = panoptic_deeplab_options.instance.low_level_override
            else:
                low_level_options = panoptic_deeplab_options.low_level

            # If instance_decoder is set, use those options; otherwise reuse the
            # architecture as defined for the semantic decoder.
            if panoptic_deeplab_options.instance.HasField(
                    'instance_decoder_override'):
                decoder_options = (panoptic_deeplab_options.instance
                                   .instance_decoder_override)

            low_level_feature_keys = [item.feature_key for item in low_level_options]
            low_level_channels_project = [
                item.channels_project for item in low_level_options
            ]

            self._instance_decoder = PanopticDeepLabSingleDecoder(
                high_level_feature_name=decoder_options.feature_key,
                low_level_feature_names=low_level_feature_keys,
                low_level_channels_project=low_level_channels_project,
                aspp_output_channels=decoder_options.aspp_channels,
                decoder_output_channels=decoder_options.decoder_channels,
                atrous_rates=decoder_options.atrous_rates,
                name='instance_decoder',
                aspp_use_only_1x1_proj_conv=(
                    decoder_options.aspp_use_only_1x1_proj_conv),
                decoder_conv_type=decoder_options.decoder_conv_type,
                bn_layer=bn_layer,
                activation=activation)
            self._instance_center_head = PanopticDeepLabSingleHead(
                panoptic_deeplab_options.instance.center_head.head_channels,
                panoptic_deeplab_options.instance.center_head.output_channels,
                common.PRED_CENTER_HEATMAP_KEY,
                name='instance_center_head',
                conv_type=(
                    panoptic_deeplab_options.instance.center_head.head_conv_type),
                bn_layer=bn_layer,
                activation=activation)
            self._instance_regression_head = PanopticDeepLabSingleHead(
                panoptic_deeplab_options.instance.regression_head.head_channels,
                panoptic_deeplab_options.instance.regression_head.output_channels,
                common.PRED_OFFSET_MAP_KEY,
                name='instance_regression_head',
                conv_type=(
                    panoptic_deeplab_options.instance.regression_head.head_conv_type),
                bn_layer=bn_layer,
                activation=activation)

    def reset_pooling_layer(self):
        """Resets the ASPP pooling layers to global average pooling."""
        self._semantic_decoder.reset_pooling_layer()
        if self._instance_decoder is not None:
            self._instance_decoder.reset_pooling_layer()

    def set_pool_size(self, pool_size):
        """Sets the pooling size of the ASPP pooling layers.

        Args:
          pool_size: A tuple specifying the pooling size of the ASPP pooling layers.
        """
        self._semantic_decoder.set_pool_size(pool_size)
        if self._instance_decoder is not None:
            self._instance_decoder.set_pool_size(pool_size)

    def get_pool_size(self):
        return self._semantic_decoder.get_pool_size()

    @property
    def checkpoint_items(self):
        items = {
            common.CKPT_SEMANTIC_DECODER:
                self._semantic_decoder,
            common.CKPT_SEMANTIC_HEAD_WITHOUT_LAST_LAYER:
                self._semantic_head.conv_block,
            common.CKPT_SEMANTIC_LAST_LAYER:
                self._semantic_head.final_conv
        }
        if self._instance_decoder is not None:
            instance_items = {
                common.CKPT_INSTANCE_DECODER:
                    self._instance_decoder,
                common.CKPT_INSTANCE_CENTER_HEAD_WITHOUT_LAST_LAYER:
                    self._instance_center_head.conv_block,
                common.CKPT_INSTANCE_CENTER_HEAD_LAST_LAYER:
                    self._instance_center_head.final_conv,
                common.CKPT_INSTANCE_REGRESSION_HEAD_WITHOUT_LAST_LAYER:
                    self._instance_regression_head.conv_block,
                common.CKPT_INSTANCE_REGRESSION_HEAD_LAST_LAYER:
                    self._instance_regression_head.final_conv,
            }
            items.update(instance_items)
        return items

    def call(self, features, training=False):
        """Performs a forward pass.

        Args:
          features: An input dict of tf.Tensor with shape [batch, height, width,
            channels]. Different keys should point to different features extracted
            by the encoder, e.g. low-level or high-level features.
          training: A boolean flag indicating whether training behavior should be
            used (default: False).

        Returns:
          A dictionary containing the results of the semantic segmentation head and
            depending on the configuration also of the instance segmentation head.
        """

        semantic_features = self._semantic_decoder(features, training=training)
        results = self._semantic_head(semantic_features, training=training)

        if self._instance_decoder is not None:
            instance_features = self._instance_decoder(features, training=training)
            instance_center_predictions = self._instance_center_head(
                instance_features, training=training)
            instance_regression_predictions = self._instance_regression_head(
                instance_features, training=training)

            if results.keys() & instance_center_predictions.keys():
                raise ValueError('The keys of the semantic branch and the instance '
                                 'center branch overlap. Please use unique keys.')
            results.update(instance_center_predictions)

            if results.keys() & instance_regression_predictions.keys():
                raise ValueError('The keys of the semantic branch and the instance '
                                 'regression branch overlap. Please use unique keys.')
            results.update(instance_regression_predictions)

        return results





def test_variable_inputs():
    """
    Test case where a node depends on a variable length set of inputs.
    """
    from geowatch.mlops.pipeline_nodes import ProcessNode
    from geowatch.mlops.pipeline_nodes import Pipeline

    # A simple pipeline where we don't need to manage reconfiguration.
    node1 = ProcessNode(name='node1', executable='node1', out_paths={'key1': 'path1'}, node_dpath='.')
    node2 = ProcessNode(name='node2', executable='node2', out_paths={'key2': 'path2'}, node_dpath='.')
    node3 = ProcessNode(name='node3', executable='node3', out_paths={'key3': 'path3'}, node_dpath='.')

    combine_node = ProcessNode(name='combine', executable='combine', in_paths={'varpaths'})

    node1.outputs['key1'].connect(combine_node.inputs['varpaths'])
    node2.outputs['key2'].connect(combine_node.inputs['varpaths'])
    node3.outputs['key3'].connect(combine_node.inputs['varpaths'])

    dag_nodes = [
        node1,
        node2,
        node3,
        combine_node
    ]
    dag = Pipeline(dag_nodes)
    dag.print_graphs()
    dag.configure()

    print(node1.final_command())
    print(node2.final_command())
    print(node3.final_command())
    print(combine_node.final_command())

    pred_nodes = combine_node.predecessor_process_nodes()
    assert list(pred_nodes) == [node1, node2, node3]


def test_slurm_options():
    from geowatch.mlops.pipeline_nodes import ProcessNode
    from geowatch.mlops.pipeline_nodes import Pipeline

    # A simple pipeline where we don't need to manage reconfiguration.
    node1 = ProcessNode(name='node1', executable='node1.exe', out_paths={'key1': 'path1'}, node_dpath='.')
    node2 = ProcessNode(name='node2', executable='node2.exe', out_paths={'key2': 'path2'}, node_dpath='.')
    node3 = ProcessNode(name='node3', executable='node3.exe', out_paths={'key3': 'path3'}, node_dpath='.')

    combine_node = ProcessNode(name='combine', executable='combine', in_paths={'varpaths'})

    node1.outputs['key1'].connect(combine_node.inputs['varpaths'])
    node2.outputs['key2'].connect(combine_node.inputs['varpaths'])
    node3.outputs['key3'].connect(combine_node.inputs['varpaths'])

    dag_nodes = [
        node1,
        node2,
        node3,
        combine_node
    ]
    dag = Pipeline(dag_nodes)
    dag.print_graphs()

    # TODO: ensure slurm options are respected in the job itself
    dag.configure(config={
        'node2.__slurm_options__': 'foo',
        'node2.some_key': 'some_value',
    })

    import cmd_queue
    cmd_queue.Queue
    queue = cmd_queue.Queue.create(backend='slurm')
    dag.submit_jobs(queue=queue)
    print_kwargs = {
        'with_status': 0,
        'style': "colors",
        'with_locks': 0,
        # 'exclude_tags': ['boilerplate'],
    }
    queue.print_commands(**print_kwargs)

    text = queue.finalize_text()
    print(text)

    new_order = queue.order_jobs()
    for job in new_order:
        print(job.__dict__)
    job = new_order[0]



# coding=utf-8
# Copyright 2022 The Deeplab2 Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This file contains functions to post-process Motion-DeepLab results."""

from typing import Tuple

import tensorflow as tf


def assign_instances_to_previous_tracks(
        prev_centers: tf.Tensor,
        current_centers: tf.Tensor,
        heatmap: tf.Tensor,
        offsets: tf.Tensor,
        panoptic_map: tf.Tensor,
        next_id: tf.Tensor,
        label_divisor: int,
        sigma=7) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
    """Greedy assignment of current centers to previous centers.

    Current centers are selected in decreasing order of confidence (heatmap
    scores). These centers are transformed with the offsets and assigned to
    previous centers.

    Args:
      prev_centers: A tf.Tensor containing previous centers of shape [Np, 5]. This
        tensor contains:
        [0]: The x-coordinate.
        [1]: The y-coordinate.
        [2]: The panoptic ID.
        [3]: The geometric mean of width and height of the instance mask.
        [4]: The number of frames that no new masks got assigned to this center.
      current_centers: A tf.Tensor containing centers of current frame of shape
        [Nc, 5]. This tensor contains:
        [0]: The x-coordinate.
        [1]: The y-coordinate.
        [2]: The panoptic ID.
        [3]: The geometric mean of width and height of the instance mask.
        [4]: The number of frames that no new masks got assigned to this center.
      heatmap: A tf.Tensor of shape [batch, height, width] containing the center
        heatmap.
      offsets: A tf.Tensor of shape [batch, height, width, 2] containing the
        center offsets.
      panoptic_map: A tf.Tensor of shape [batch, height, width] containing the
        panoptic segmentation.
      next_id: A tf.Tensor of shape [1] containing the next ID.
      label_divisor: An integer specifying the label divisor for panoptic IDs.
      sigma: An optional integer specifying the number of frames that unmatched
        centers should be kept (default: 7).

    Returns:
      A tuple of three tf.Tensor:
        1. The updated panoptic segmentation map that contains track IDs.
        2. The updated tensor containing all current centers (including unmatched
          previous ones).
        3. The updated next ID that can be used for new tracks.
    """
    # Switch x and y coordinates for indexing.
    center_indices = tf.concat(
        [tf.zeros([tf.shape(current_centers)[0], 1], dtype=tf.int32),
         current_centers[:, 1:2], current_centers[:, 0:1]],
        axis=1)
    confidence_scores = tf.gather_nd(heatmap, center_indices)

    scores = tf.argsort(confidence_scores, direction='DESCENDING')

    def cond(i, *_):
        return i < tf.shape(center_indices)[0]

    def body(i, current_centers_loop, prev_centers_loop, new_panoptic_map_loop,
             next_id_loop):
        row_index = scores[i]
        i = tf.add(i, 1)
        center_id = current_centers_loop[row_index, 2]
        center_location = current_centers_loop[row_index, :2]
        center_offset_yx = offsets[0, center_location[1], center_location[0], :]
        center_offset_xy = center_offset_yx[::-1]
        center_location = center_offset_xy + tf.cast(center_location, tf.float32)
        center_sem_id = center_id // label_divisor
        center_mask = tf.equal(panoptic_map, center_id)
        prev_centers_class = prev_centers_loop[:, 2] // label_divisor
        prev_centers_with_same_class = tf.squeeze(
            tf.cast(
                tf.gather(
                    prev_centers_loop,
                    tf.where(tf.equal(prev_centers_class, center_sem_id)),
                    axis=0), tf.float32),
            axis=1)

        # Check if there are still unassigned previous centers of the same class.
        if tf.shape(prev_centers_with_same_class)[0] > 0:
            # For efficieny reasons, we do not take the sqrt when we compute the
            # minimal distances. See render_panoptic_map_as_heatmap as well.
            distances = tf.reduce_sum(
                tf.square(prev_centers_with_same_class[:, :2] - center_location),
                axis=1)
            prev_center_index = tf.math.argmin(
                distances, axis=0, output_type=tf.int32)
            min_dist = distances[prev_center_index]

            # If previous center is within a certain range, continue track.
            if min_dist < prev_centers_with_same_class[prev_center_index, 3]:
                new_center_id = tf.cast(
                    prev_centers_with_same_class[prev_center_index, 2], dtype=tf.int32)
                shape = new_panoptic_map_loop.get_shape()
                new_panoptic_map_loop = tf.where(center_mask, new_center_id,
                                                 new_panoptic_map_loop)
                new_panoptic_map_loop.set_shape(shape)
                current_centers_loop = tf.tensor_scatter_nd_update(
                    current_centers_loop, tf.expand_dims([row_index, 2], 0),
                    [new_center_id])
                # Remove previous center.
                prev_centers_loop = tf.squeeze(
                    tf.gather(
                        prev_centers_loop,
                        tf.where(tf.not_equal(prev_centers_loop[:, 2], new_center_id)),
                        axis=0),
                    axis=1)
                return (i, current_centers_loop, prev_centers_loop,
                        new_panoptic_map_loop, next_id_loop)
            else:
                # Assign new track ID
                new_center_id = center_sem_id * label_divisor + next_id_loop
                shape = new_panoptic_map_loop.get_shape()
                new_panoptic_map_loop = tf.where(center_mask, new_center_id,
                                                 new_panoptic_map_loop)
                new_panoptic_map_loop.set_shape(shape)
                current_centers_loop = tf.tensor_scatter_nd_update(
                    current_centers_loop, tf.expand_dims([row_index, 2], 0),
                    [new_center_id])
                next_id_loop += 1
                return (i, current_centers_loop, prev_centers_loop,
                        new_panoptic_map_loop, next_id_loop)
        else:
            # Assign new track ID
            new_center_id = center_sem_id * label_divisor + next_id_loop
            shape = new_panoptic_map_loop.get_shape()
            new_panoptic_map_loop = tf.where(center_mask, new_center_id,
                                             new_panoptic_map_loop)
            new_panoptic_map_loop.set_shape(shape)
            current_centers_loop = tf.tensor_scatter_nd_update(
                current_centers_loop, tf.expand_dims([row_index, 2], 0),
                [new_center_id])
            next_id_loop += 1
            return (i, current_centers_loop, prev_centers_loop, new_panoptic_map_loop,
                    next_id_loop)

    loop_start_index = tf.constant(0)
    (_, current_centers,
     unmatched_centers, new_panoptic_map, next_id) = tf.while_loop(
        cond, body,
        (loop_start_index, current_centers, prev_centers, panoptic_map,
         next_id))

    # Keep unmatched centers for sigma frames.
    if tf.shape(unmatched_centers)[0] > 0:
        current_centers = tf.concat([current_centers, unmatched_centers], axis=0)

    number_centers = tf.shape(current_centers)[0]
    indices_row = tf.range(number_centers, dtype=tf.int32)
    indices_column = tf.repeat([4], number_centers, axis=0)
    indices = tf.stack([indices_row, indices_column], axis=1)
    current_centers = tf.tensor_scatter_nd_add(
        current_centers, indices,
        tf.repeat([1], number_centers, axis=0))

    # Remove centers after sigma frames.
    current_centers = tf.squeeze(
        tf.gather(
            current_centers,
            tf.where(tf.not_equal(current_centers[:, 4], sigma)),
            axis=0),
        axis=1)

    return new_panoptic_map, current_centers, next_id


def render_panoptic_map_as_heatmap(
        panoptic_map: tf.Tensor, sigma: int, label_divisor: int,
        void_label: int) -> Tuple[tf.Tensor, tf.Tensor]:
    """Extracts centers from panoptic map and renders as heatmap."""
    gaussian_size = 6 * sigma + 3
    x = tf.range(gaussian_size, dtype=tf.float32)
    y = tf.expand_dims(x, axis=1)
    x0, y0 = 3 * sigma + 1, 3 * sigma + 1
    gaussian = tf.math.exp(-((x - x0) ** 2 + (y - y0) ** 2) / (2 * sigma ** 2))
    gaussian = tf.cast(tf.reshape(gaussian, [-1]), tf.float32)

    height = tf.shape(panoptic_map)[1]
    width = tf.shape(panoptic_map)[2]
    # Pad center to make boundary handling easier.
    center_pad_begin = int(round(3 * sigma + 1))
    center_pad_end = int(round(3 * sigma + 2))
    center_pad = center_pad_begin + center_pad_end

    center = tf.zeros((height + center_pad, width + center_pad))
    unique_ids, _ = tf.unique(tf.reshape(panoptic_map, [-1]))
    centers_and_ids = tf.TensorArray(
        tf.int32, size=0, dynamic_size=True, clear_after_read=False)
    counter = tf.zeros([], dtype=tf.int32)

    for panoptic_id in unique_ids:
        semantic_id = panoptic_id // label_divisor
        # Filter out IDs that should be ignored, are stuff classes or crowd.
        # Stuff classes and crowd regions both have IDs of the form panoptic_id =
        # semantic_id * label_divisor
        if semantic_id == void_label or panoptic_id % label_divisor == 0:
            continue

        # Convert [[0, y0, x0], ...] to [[0, ...], [y0, ...], [x0, ...]].
        mask_index = tf.cast(
            tf.transpose(tf.where(panoptic_map == panoptic_id)), tf.float32)
        mask_size = (
                tf.reduce_max(mask_index, axis=1) - tf.reduce_min(mask_index, axis=1))
        # The radius is defined as the geometric mean of width and height.
        # For efficieny reasons, we do not take the sqrt when we compute the minimal
        # distances. See assign_instances_to_previous_tracks as well.
        mask_radius = tf.cast(tf.round(mask_size[1] * mask_size[2]), tf.int32)
        centers = tf.reduce_mean(mask_index, axis=1)

        center_x = tf.cast(tf.round(centers[2]), tf.int32)
        center_y = tf.cast(tf.round(centers[1]), tf.int32)
        centers_and_ids = centers_and_ids.write(
            counter,
            [center_x, center_y, tf.cast(panoptic_id, tf.int32), mask_radius, 0])
        counter += 1

        # Due to the padding with center_pad_begin in center, the computed center
        # becomes the upper left corner in the center tensor.
        upper_left = center_x, center_y
        bottom_right = (upper_left[0] + gaussian_size,
                        upper_left[1] + gaussian_size)

        indices_x, indices_y = tf.meshgrid(
            tf.range(upper_left[0], bottom_right[0]),
            tf.range(upper_left[1], bottom_right[1]))
        indices = tf.transpose(
            tf.stack([tf.reshape(indices_y, [-1]),
                      tf.reshape(indices_x, [-1])]))

        center = tf.tensor_scatter_nd_max(
            center, indices, gaussian, name='center_scatter')

    center = center[center_pad_begin:(center_pad_begin + height), center_pad_begin:(center_pad_begin + width)]
    return tf.expand_dims(center, axis=0), centers_and_ids.stack()



# mypy: ignore-errors
from __future__ import annotations

from typing import Union, Any, Callable
from collections import Counter
from contextlib import nullcontext

from graphviz import Digraph
from torch.nn.modules import Identity

from .computation_node import NodeContainer
from .computation_node import TensorNode, ModuleNode, FunctionNode
from .utils import updated_dict, assert_input_type

COMPUTATION_NODES = Union[TensorNode, ModuleNode, FunctionNode]

node2color = {
    TensorNode: "lightyellow",
    ModuleNode: "darkseagreen1",
    FunctionNode: "aliceblue",
}

# TODO: Currently, we only use directed graphviz graph since DNN are
# graphs except for e.g. graph neural network (GNN). Experiment on GNN
# and see if undirected graphviz graph can be used to represent GNNs

# TODO: change api of to include also function calls, not only pytorch models
# so, keep the api here as general as possible


class ComputationGraph:
    '''A class to represent Computational graph and visualization of pytorch model

    Attributes:
        visual_graph (Digraph):
            Graphviz.Digraph object to represent computational graph of
            pytorch model

        root_container (NodeContainer):
            Iterable of TensorNodes to represent all input/root nodes
            of pytorch model.

        show_shapes (bool):
            Whether to show shapes of tensor/input/outputs

        hide_module_functions (bool):
            Some modules contain only torch.function and no submodule,
            e.g. nn.Conv2d. They are usually implemented to do one type
            of computation, e.g. Conv2d -> 2D Convolution. If True,
            visual graph only displays the module itself,
            while ignoring its inner functions.

        hide_inner_tensors (bool):
            Whether to hide inner tensors in graphviz graph object


        node_hierarchy dict:
            Represents nested hierarchy of ComputationNodes by nested dictionary
    '''
    def __init__(
        self,
        visual_graph: Digraph,
        root_container: NodeContainer[TensorNode],
        show_shapes: bool = True,
        expand_nested: bool = False,
        hide_inner_tensors: bool = True,
        hide_module_functions: bool = True,
        roll: bool = True,
        depth: int | float = 3,
    ):
        '''
        Resets the running_node_id, id_dict when a new ComputationGraph is initialized.
        Otherwise, labels would depend on previous ComputationGraph runs
        '''
        self.visual_graph = visual_graph
        self.root_container = root_container
        self.show_shapes = show_shapes
        self.expand_nested = expand_nested
        self.hide_inner_tensors = hide_inner_tensors
        self.hide_module_functions = hide_module_functions
        self.roll = roll
        self.depth = depth

        # specs for html table, needed for node labels
        self.html_config = {
            'border': 0,
            'cell_border': 1,
            'cell_spacing': 0,
            'cell_padding': 4,
            'col_span': 2,
            'row_span': 2,
        }
        self.reset_graph_history()

    def reset_graph_history(self) -> None:
        '''Resets to id config to the setting of empty visual graph
        needed for getting reproducible/deterministic node name and
        graphviz graphs. This is especially important for output tests
        '''
        self.context_tracker = {'current_context': [], 'current_depth': 0}
        self.running_node_id: int = 0
        self.running_subgraph_id: int = 0
        self.id_dict: dict[str, int] = {}
        self.node_set: set[int] = set()
        self.edge_list: list[tuple[COMPUTATION_NODES, COMPUTATION_NODES]] = []

        # module node  to capture whole graph
        main_container_module = ModuleNode(Identity(), -1)
        main_container_module.is_container = False
        self.subgraph_dict: dict[str, int] = {main_container_module.node_id: 0}
        self.running_subgraph_id += 1

        # Add input nodes
        self.node_hierarchy = {
            main_container_module: list(self.root_container)
        }
        for root_node in self.root_container:
            root_node.context = self.node_hierarchy[main_container_module]

    def fill_visual_graph(self) -> None:
        '''Fills the graphviz graph with desired nodes and edges.'''

        self.render_nodes()
        self.render_edges()
        self.resize_graph()

    def render_nodes(self) -> None:
        kwargs = {
            'cur_node': self.node_hierarchy,
            'subgraph': None,
        }
        self.traverse_graph(self.collect_graph, **kwargs)

    def render_edges(self) -> None:
        '''Records all edges in self.edge_list to
        the graphviz graph using node ids from edge_list'''
        edge_counter: dict[tuple[int, int], int] = {}
        for tail, head in self.edge_list:
            edge_id = self.id_dict[tail.node_id], self.id_dict[head.node_id]
            edge_counter[edge_id] = edge_counter.get(edge_id, 0) + 1
            self.add_edge(edge_id, edge_counter[edge_id])

    def traverse_graph(
        self, action_fn: Callable[..., None], **kwargs: Any
    ) -> None:
        cur_node = kwargs['cur_node']
        cur_subgraph = (
            self.visual_graph if kwargs['subgraph'] is None else kwargs['subgraph']
        )
        assert_input_type(
            'traverse_graph', (TensorNode, ModuleNode, FunctionNode, dict), cur_node
        )
        if isinstance(cur_node, (TensorNode, ModuleNode, FunctionNode)):
            if cur_node.depth <= self.depth:
                action_fn(**kwargs)
            return

        if isinstance(cur_node, dict):
            k, v = list(cur_node.items())[0]
            new_kwargs = updated_dict(kwargs, 'cur_node', k)
            if k.depth <= self.depth and k.depth >= 0:
                action_fn(**new_kwargs)

            # if it is container module, move directly to outputs
            if self.hide_module_functions and k.is_container:
                for g in k.output_nodes:
                    new_kwargs = updated_dict(new_kwargs, 'cur_node', g)
                    self.traverse_graph(action_fn, **new_kwargs)
                return

            display_nested = (
                k.depth < self.depth and k.depth >= 1 and self.expand_nested
            )

            with (
                cur_subgraph.subgraph(name=f'cluster_{self.subgraph_dict[k.node_id]}')
                if display_nested else nullcontext()
            ) as cur_cont:
                if display_nested:
                    cur_cont.attr(
                        style='dashed', label=k.name, labeljust='l', fontsize='12'
                    )
                    new_kwargs = updated_dict(new_kwargs, 'subgraph', cur_cont)
                for g in v:
                    new_kwargs = updated_dict(new_kwargs, 'cur_node', g)
                    self.traverse_graph(action_fn, **new_kwargs)

    def collect_graph(self, **kwargs: Any) -> None:
        '''Adds edges and nodes with appropriate node name/id (so it respects
        properties e.g. if rolled recursive nodes are given the same node name
        in graphviz graph)'''

        cur_node = kwargs['cur_node']
        # if tensor node is traced, dont repeat collecting
        # if node is isolated, dont record it
        is_isolated = cur_node.is_root() and cur_node.is_leaf()
        if id(cur_node) in self.node_set or is_isolated:
            return

        self.check_node(cur_node)
        is_cur_visible = self.is_node_visible(cur_node)
        # add node
        if is_cur_visible:
            subgraph = kwargs['subgraph']
            if isinstance(cur_node, (FunctionNode, ModuleNode)):
                if self.roll:
                    self.rollify(cur_node)
                self.add_node(cur_node, subgraph)

            if isinstance(cur_node, TensorNode):
                self.add_node(cur_node, subgraph)

        elif isinstance(cur_node, ModuleNode):
            # add subgraph
            if self.roll:
                self.rollify(cur_node)
            if cur_node.node_id not in self.subgraph_dict:
                self.subgraph_dict[cur_node.node_id] = self.running_subgraph_id
                self.running_subgraph_id += 1

        # add edges only through
        # node -> TensorNode -> Node connection
        if not isinstance(cur_node, TensorNode):
            return

        # add edges
        # {cur_node -> head} part
        tail_node = self.get_tail_node(cur_node)
        is_main_node_visible = self.is_node_visible(cur_node.main_node)
        is_tail_node_visible = self.is_node_visible(tail_node)
        if not cur_node.is_leaf():
            for children_node in cur_node.children:
                is_output_visible = self.is_node_visible(children_node)
                if is_output_visible:
                    if is_main_node_visible:
                        self.edge_list.append((cur_node, children_node))
                    elif is_tail_node_visible:
                        self.edge_list.append((tail_node, children_node))

        # {tail -> cur_node} part
        # # output node
        # visible tensor and non-input tensor nodes
        if is_cur_visible and not cur_node.is_root():
            assert not isinstance(tail_node, TensorNode) or tail_node.is_root(), (
                "get_tail_node function returned inconsistent Node, please report this"
            )
            self.edge_list.append((tail_node, cur_node))

    def rollify(self, cur_node: ModuleNode | FunctionNode) -> None:
        '''Rolls computational graph by identifying recursively used
        Modules. This is done by giving the same id for nodes that are
        recursively used.
        This becomes complex when there are stateless and torch.functions.
        For more details see docs'''

        head_node = next(iter(cur_node.output_nodes))
        if not head_node.is_leaf() and self.hide_inner_tensors:
            head_node = next(iter(head_node.children))

        # identify recursively used modules
        # with the same node id
        output_id = get_output_id(head_node)
        cur_node.set_node_id(output_id=output_id)

    def is_node_visible(self, compute_node: COMPUTATION_NODES) -> bool:
        '''Returns True if node should be displayed on the visual
        graph. Otherwise False'''

        assert_input_type(
            'is_node_visible', (TensorNode, ModuleNode, FunctionNode,), compute_node
        )

        if compute_node.name == 'empty-pass':
            return False

        if isinstance(compute_node, (ModuleNode, FunctionNode)):
            is_visible = (
                isinstance(compute_node, FunctionNode) or (
                    (self.hide_module_functions and compute_node.is_container)
                    or compute_node.depth == self.depth
                )
            )
            return is_visible

        else:
            if compute_node.main_node.depth < 0 or compute_node.is_aux:
                return False

            is_main_input_or_output = (
                (compute_node.is_root() or compute_node.is_leaf())
                and compute_node.depth == 0
            )
            is_visible = (
                not self.hide_inner_tensors or is_main_input_or_output
            )

            return is_visible

    def get_tail_node(self, _tensor_node: TensorNode) -> COMPUTATION_NODES:

        tensor_node = _tensor_node.main_node if _tensor_node.is_aux else _tensor_node

        # non-output nodes eminating from input node
        if tensor_node.is_root():
            return tensor_node

        current_parent_h = tensor_node.parent_hierarchy

        sorted_depth = sorted(depth for depth in current_parent_h)
        tail_node = next(iter(tensor_node.parents))
        depth = 0
        for depth in sorted_depth:
            tail_node = current_parent_h[depth]
            if depth >= self.depth:
                break

        module_depth = depth - 1
        # if returned by container module and hide_module_functions
        if (
            isinstance(current_parent_h[depth], FunctionNode) and
            module_depth in tensor_node.parent_hierarchy and self.hide_module_functions
        ):
            if current_parent_h[module_depth].is_container:
                return current_parent_h[module_depth]

        # Even though this is recursive, not harmful for complexity
        # The reason: the (time) complexity ~ O(L^2) where L
        # is the length of CONTINUOUS path along which the same tensor is passed
        # without any operation on it. L is always small since we dont use
        # infinitely big network with infinitely big continuou pass of unchanged
        # tensor. This recursion is necessary e.g. for LDC model
        if tail_node.name == 'empty-pass':
            empty_pass_parent = next(iter((tail_node.parents)))
            assert isinstance(empty_pass_parent, TensorNode), (
                f'{empty_pass_parent} is input of {tail_node}'
                f'and must a be TensorNode'
            )
            return self.get_tail_node(empty_pass_parent)
        return tail_node

    def add_edge(
        self, edge_ids: tuple[int, int], edg_cnt: int
    ) -> None:

        tail_id, head_id = edge_ids
        label = None if edg_cnt == 1 else f' x{edg_cnt}'
        self.visual_graph.edge(f'{tail_id}', f'{head_id}', label=label)

    def add_node(
        self, node: COMPUTATION_NODES, subgraph: Digraph | None = None
    ) -> None:
        '''Adds node to the graphviz with correct id, label and color
        settings. Updates state of running_node_id if node is not
        identified before.'''
        if node.node_id not in self.id_dict:
            self.id_dict[node.node_id] = self.running_node_id
            self.running_node_id += 1
        label = self.get_node_label(node)
        node_color = ComputationGraph.get_node_color(node)

        if subgraph is None:
            subgraph = self.visual_graph
        subgraph.node(
            name=f'{self.id_dict[node.node_id]}', label=label, fillcolor=node_color,
        )
        self.node_set.add(id(node))

    def get_node_label(self, node: COMPUTATION_NODES) -> str:
        '''Returns html-like format for the label of node. This html-like
        label is based on Graphviz API for html-like format. For setting of node label
        it uses graph config and html_config.'''
        input_str = 'input'
        output_str = 'output'
        border = self.html_config['border']
        cell_sp = self.html_config['cell_spacing']
        cell_pad = self.html_config['cell_padding']
        cell_bor = self.html_config['cell_border']
        if self.show_shapes:
            if isinstance(node, TensorNode):
                label = f'''<
                    <TABLE BORDER="{border}" CELLBORDER="{cell_bor}"
                    CELLSPACING="{cell_sp}" CELLPADDING="{cell_pad}">
                        <TR><TD>{node.name}<BR/>depth:{node.depth}</TD><TD>{node.tensor_shape}</TD></TR>
                    </TABLE>>'''
            else:
                input_repr = compact_list_repr(node.input_shape)
                output_repr = compact_list_repr(node.output_shape)
                label = f'''<
                    <TABLE BORDER="{border}" CELLBORDER="{cell_bor}"
                    CELLSPACING="{cell_sp}" CELLPADDING="{cell_pad}">
                    <TR>
                        <TD ROWSPAN="2">{node.name}<BR/>depth:{node.depth}</TD>
                        <TD COLSPAN="2">{input_str}:</TD>
                        <TD COLSPAN="2">{input_repr} </TD>
                    </TR>
                    <TR>
                        <TD COLSPAN="2">{output_str}: </TD>
                        <TD COLSPAN="2">{output_repr} </TD>
                    </TR>
                    </TABLE>>'''
        else:
            label = f'''<
                    <TABLE BORDER="{border}" CELLBORDER="{cell_bor}"
                    CELLSPACING="{cell_sp}" CELLPADDING="{cell_pad}">
                        <TR><TD>{node.name}<BR/>depth:{node.depth}</TD></TR>
                    </TABLE>>'''
        return label

    def resize_graph(
        self,
        scale: float = 1.0,
        size_per_element: float = 0.3,
        min_size: float = 12
    ) -> None:
        """Resize the graph according to how much content it contains.
        Modify the graph in place. Default values are subject to change,
        so far they seem to work fine.
        """
        # Get the approximate number of nodes and edges
        num_rows = len(self.visual_graph.body)
        content_size = num_rows * size_per_element
        size = scale * max(min_size, content_size)
        size_str = str(size) + "," + str(size)
        self.visual_graph.graph_attr.update(size=size_str,)

    @staticmethod
    def get_node_color(
        node: COMPUTATION_NODES
    ) -> str:
        return node2color[type(node)]

    def check_node(self, node: COMPUTATION_NODES) -> None:
        assert node.node_id != 'null', f'wrong id {node} {type(node)}'
        assert '-' not in node.node_id, 'No repetition of node recording is allowed'
        assert node.depth <= self.depth, f"Exceeds display depth limit, {node}"
        assert (
            sum(1 for _ in node.parents) in [0, 1] or not isinstance(node, TensorNode)
        ), (
            f'tensor must have single input node {node}'
        )


def compact_list_repr(x: list[Any]) -> str:
    '''returns more compact representation of list with
    repeated elements. This is useful for e.g. output of transformer/rnn
    models where hidden state outputs shapes is repetation of one hidden unit
    output'''

    list_counter = Counter(x)
    x_repr = ''

    for elem, cnt in list_counter.items():
        if cnt == 1:
            x_repr += f'{elem}, '
        else:
            x_repr += f'{cnt} x {elem}, '

    # get rid of last comma
    return x_repr[:-2]


def get_output_id(head_node: COMPUTATION_NODES) -> str:
    ''' This returns id of output to get correct id.
    This is used to identify the recursively used modules.
    Identification relation is as follows:
        ModuleNodes => by id of nn.Module object
        Parameterless ModulesNodes => by id of nn.Module object
        FunctionNodes => by id of Node object
    '''
    if isinstance(head_node, ModuleNode):
        output_id = str(head_node.compute_unit_id)
    else:
        output_id = head_node.node_id

    return output_id



"""
Checks that the scheduler builds appropriate commands.
"""


def test_bas_pipline_schedule_default_params():
    from geowatch.mlops import schedule_evaluation
    import ubelt as ub
    dpath = ub.Path.appdir('geowatch/unit_tests/scheduler/unit0').ensuredir()

    dvc_data_dpath = ub.Path('/my_data_dvc')

    config = schedule_evaluation.ScheduleEvaluationConfig(**{
        'run': 0,
        'root_dpath': dpath,
        'pipeline': 'bas',
        'backend': 'serial',
        'enable_links': False,
        'params': ub.codeblock(
            f'''
            bas_poly_eval.true_site_dpath: {dvc_data_dpath}/annotations/site_models
            bas_poly_eval.true_region_dpath: {dvc_data_dpath}/annotations/region_models
            '''
        )
    })
    dag, queue = schedule_evaluation.schedule_evaluation(config)

    bas_poly_job = None
    bas_poly_eval_job = None
    for job in queue.jobs:
        if job.name.startswith('bas_poly_id'):
            bas_poly_job = job
        if job.name.startswith('bas_poly_eval_id'):
            bas_poly_eval_job = job

    assert bas_poly_job is not None
    assert bas_poly_eval_job is not None
    assert "--boundary_region=None" in bas_poly_job.command


def test_bas_pipline_schedule1():
    from geowatch.mlops import schedule_evaluation
    import ubelt as ub
    dpath = ub.Path.appdir('geowatch/unit_tests/scheduler/unit1').ensuredir()

    dvc_data_dpath = ub.Path('/my_data_dvc')
    dvc_expt_dpath = ub.Path('/my_expt_dvc')

    config = schedule_evaluation.ScheduleEvaluationConfig(**{
        'run': 0,
        'root_dpath': dpath,
        'pipeline': 'bas',
        'backend': 'serial',
        'enable_links': False,
        'params': ub.codeblock(
            f'''
            bas_pxl.package_fpath:
                - {dvc_expt_dpath}/models/fusion/Drop4-BAS/packages/Drop4_TuneV323_BAS_30GSD_BGRNSH_V2/package_epoch0_step41.pt.pt
            bas_pxl.test_dataset:
                - {dvc_data_dpath}/Drop6_MeanYear/imganns-KR_R001.kwcoco.zip
            bas_poly.boundary_region: {dvc_data_dpath}/annotations/region_models
            bas_poly_eval.true_site_dpath: {dvc_data_dpath}/annotations/site_models
            bas_poly_eval.true_region_dpath: {dvc_data_dpath}/annotations/region_models
            '''
        )
    })
    dag, queue = schedule_evaluation.schedule_evaluation(config)

    bas_poly_job = None
    bas_poly_eval_job = None
    for job in queue.jobs:
        if job.name.startswith('bas_poly_id'):
            bas_poly_job = job
        if job.name.startswith('bas_poly_eval_id'):
            bas_poly_eval_job = job

    assert bas_poly_job is not None
    assert bas_poly_eval_job is not None
    assert "--boundary_region 'None'" not in bas_poly_job.command


def test_joint_bas_sc_pipline_schedule1():
    from geowatch.mlops import schedule_evaluation
    import ubelt as ub
    dpath = ub.Path.appdir('geowatch/unit_tests/scheduler/unit2').ensuredir()

    dvc_data_dpath = ub.Path('/my_data_dvc')
    dvc_expt_dpath = ub.Path('/my_expt_dvc')

    config = schedule_evaluation.ScheduleEvaluationConfig(**{
        'run': 0,
        'root_dpath': dpath,
        'pipeline': 'joint_bas_sc',
        'backend': 'serial',
        'enable_links': False,
        'params': ub.codeblock(
            f'''
            bas_pxl.package_fpath:
                - {dvc_expt_dpath}/models/fusion/Drop4-BAS/packages/Drop4_TuneV323_BAS_30GSD_BGRNSH_V2/package_epoch0_step41.pt.pt
            bas_pxl.test_dataset:
                - {dvc_data_dpath}/Drop6_MeanYear/imganns-KR_R001.kwcoco.zip
            bas_poly.boundary_region: {dvc_data_dpath}/annotations/region_models
            bas_poly_eval.true_site_dpath: {dvc_data_dpath}/annotations/site_models
            bas_poly_eval.true_region_dpath: {dvc_data_dpath}/annotations/region_models
            sc_poly_eval.true_site_dpath: {dvc_data_dpath}/annotations/site_models
            sc_poly_eval.true_region_dpath: {dvc_data_dpath}/annotations/region_models
            '''
        )
    })
    dag, queue = schedule_evaluation.schedule_evaluation(config)

    bas_poly_job = None
    bas_poly_eval_job = None
    for job in queue.jobs:
        if job.name.startswith('bas_poly_id'):
            bas_poly_job = job
        if job.name.startswith('bas_poly_eval_id'):
            bas_poly_eval_job = job

    assert bas_poly_job is not None
    assert bas_poly_eval_job is not None
    assert "--boundary_region 'None'" not in bas_poly_job.command


def demodata_pipeline(dpath):
    import ubelt as ub
    script_fpath = dpath / 'script.py'
    pipeline_fpath = dpath / '_simple_demo_pipeline_v003.py'

    script_text = ub.codeblock(
        '''
        #!/usr/bin/env python3
        import scriptconfig as scfg
        import ubelt as ub
        import json


        class ScriptCLI(scfg.DataConfig):
            src = 'input.json'
            dst = 'output.json'
            param1 = None
            param2 = None
            param3 = None

            @classmethod
            def main(cls, cmdline=1, **kwargs):
                import rich
                from rich.markup import escape
                config = cls.cli(cmdline=cmdline, data=kwargs, strict=True)
                rich.print('config = ' + escape(ub.urepr(config, nl=1)))
                src_fpath = ub.Path(config.src)
                dst_fpath = ub.Path(config.dst)
                src_text = src_fpath.read_text()
                src_data = json.loads(src_text)

                hidden = int(ub.hash_data([config.param1, config.param2, config.param3], base=10, hasher='sha1'))
                flags = [c == '1' for c in bin(hidden)[2:]]
                goodness = sum(flags) / len(flags)

                dst_data = {'size': len(src_text), 'goodness': goodness, 'nest': src_data}
                dst_fpath.parent.ensuredir()
                dst_fpath.write_text(json.dumps(dst_data))

        __cli__ = ScriptCLI

        if __name__ == '__main__':
            __cli__.main()
        ''')
    # Test the code compiles and write it to disk
    compile(script_text, mode='exec', filename='<test-compile>')
    script_fpath.write_text(script_text)

    pipeline_text = ub.codeblock(
        '''
        from geowatch.mlops.pipeline_nodes import ProcessNode
        from geowatch.mlops.pipeline_nodes import PipelineDAG

        class Step1(ProcessNode):
            name = 'step1'
            executable = 'python ''' + str(script_fpath) + ''''
            in_paths = {
                'src',
            }
            out_paths = {
                'dst': 'step1_output.json',
            }

            def load_result(self, node_dpath):
                import json
                from geowatch.mlops.aggregate_loader import new_process_context_parser
                from geowatch.utils import util_dotdict
                fpath = node_dpath / self.out_paths[self.primary_out_key]
                data = json.loads(fpath.read_text())
                nest_resolved = {}
                nest_resolved['metrics.size'] = data['size']
                nest_resolved['metrics.goodness'] = data['goodness']
                flat_resolved = util_dotdict.DotDict.from_nested(nest_resolved)
                flat_resolved = flat_resolved.insert_prefix(self.name, index=1)
                return flat_resolved

        def build_pipeline():
            nodes = {}
            nodes['step1'] = Step1()
            dag = PipelineDAG(nodes)
            dag.build_nx_graphs()
            return dag
        ''')

    # Test that the code compiles
    compile(pipeline_text, mode='exec', filename='<test-compile>')
    pipeline_fpath.write_text(pipeline_text)
    return pipeline_fpath


def test_simple_slurm_dry_run():
    """
    Ignore:
        python ~/code/geowatch/tests/test_mlops_scheduler.py test_simple_but_real_custom_pipeline

    Ignore:
        import sys, ubelt
        sys.path.append(ubelt.expandpath('~/code/geowatch/tests'))
        from test_mlops_scheduler import *  # NOQA
    """
    from geowatch.mlops import schedule_evaluation
    import ubelt as ub
    dpath = ub.Path.appdir('geowatch/unit_tests/scheduler/test_slurm_dryrun').ensuredir()

    pipeline_fpath = demodata_pipeline(dpath)

    input_fpath = dpath / 'input.json'
    input_fpath.write_text('{"type": "orig_input"}')

    root_dpath = (dpath / 'runs').delete().ensuredir()
    config = schedule_evaluation.ScheduleEvaluationConfig(**{
        'run': 0,
        'root_dpath': root_dpath,
        'backend': 'slurm',
        'params': ub.codeblock(
            f'''
            pipeline: {pipeline_fpath}::build_pipeline()
            matrix:
                step1.src:
                    - {input_fpath}
                step1.param1: |
                    - this: "is text 100% representing"
                      some: "yaml config"
                      omg: "single ' quote"
                      eek: 'double " quote'
                step1.param2:
                    - option1
                    - option2
                step1.param3:
                    - 4.5
                    - 9.2
                    - 3.14159
                    - 2.71828
            '''
        )
    })

    print('Dry run first')
    config['run'] = 0
    dag, queue = schedule_evaluation.schedule_evaluation(config)


def test_simple_but_real_custom_pipeline():
    """
    Ignore:
        python ~/code/geowatch/tests/test_mlops_scheduler.py test_simple_but_real_custom_pipeline

    Ignore:
        import sys, ubelt
        sys.path.append(ubelt.expandpath('~/code/geowatch/tests'))
        from test_mlops_scheduler import *  # NOQA
    """
    from geowatch.mlops import schedule_evaluation
    from geowatch.mlops import aggregate
    import ubelt as ub
    dpath = ub.Path.appdir('geowatch/unit_tests/scheduler/test_real_pipeline').ensuredir()

    pipeline_fpath = demodata_pipeline(dpath)

    input_fpath = dpath / 'input.json'
    input_fpath.write_text('{"type": "orig_input"}')

    root_dpath = (dpath / 'runs').delete().ensuredir()
    config = schedule_evaluation.ScheduleEvaluationConfig(**{
        'run': 0,
        'root_dpath': root_dpath,
        'backend': 'serial',
        'params': ub.codeblock(
            f'''
            pipeline: {pipeline_fpath}::build_pipeline()
            matrix:
                step1.src:
                    - {input_fpath}
                step1.param1: |
                    - this: "is text 100% representing"
                      some: "yaml config"
                      omg: "single ' quote"
                      eek: 'double " quote'
                step1.param2:
                    - option1
                    - option2
                step1.param3:
                    - 4.5
                    - 9.2
                    - 3.14159
                    - 2.71828
            '''
        )
    })

    print('Dry run first')
    config['run'] = 0
    dag, queue = schedule_evaluation.schedule_evaluation(config)

    print('Real run second')
    config['run'] = 1
    dag, queue = schedule_evaluation.schedule_evaluation(config)

    # Test that all job config files are readable
    import json
    for job_config_fpath in dag.root_dpath.glob('flat/step1/*/job_config.json'):
        config = json.loads(job_config_fpath.read_text())

    # Can we test that this is well formatted?
    for invoke_fpath in dag.root_dpath.glob('flat/step1/*/invoke.sh'):
        command = invoke_fpath.read_text()
        command

    agg_config = aggregate.AggregateEvluationConfig(
        target=root_dpath,
        pipeline=f'{pipeline_fpath}::build_pipeline()',
        output_dpath=(root_dpath / 'aggregate'),
        io_workers=0,
        eval_nodes=['step1'],
    )
    eval_type_to_aggregator = aggregate.run_aggregate(agg_config)
    agg = eval_type_to_aggregator['step1']
    print(f'agg = {ub.urepr(agg, nl=1)}')


if __name__ == '__main__':
    """
    CommandLine:
        python ~/code/geowatch/tests/test_mlops_scheduler.py
    """
    import xdoctest
    xdoctest.doctest_module(__file__)
